# -*- coding: UTF-8 -*-
"""
TNN的相关函数。
包括时空代数的基本算子等。
"""


import torch
import torch.nn as nn
import matterhorn_pytorch.snn.functional as _SF
import matterhorn_pytorch.data.functional as _DF
from typing import Any as _Any


"""
以下是根据时空代数原语完备性定义的算子。
"""
op_min = lambda x, y, min, max, inh: min(x, y)
op_xmin = lambda x, y, min, max, inh: min(op_lt(x, y, min, max, inh), op_lt(y, x, min, max, inh))
op_max = lambda x, y, min, max, inh: max(x, y)
op_xmax = lambda x, y, min, max, inh: min(op_gt(x, y, min, max, inh), op_gt(y, x, min, max, inh))
op_eq = lambda x, y, min, max, inh: max(op_le(x, y, min, max, inh), op_le(y, x, min, max, inh))
op_ne = lambda x, y, min, max, inh: min(op_lt(x, y, min, max, inh), op_gt(x, y, min, max, inh))
op_lt = lambda x, y, min, max, inh: inh(x, y)
op_le = lambda x, y, min, max, inh: inh(x, inh(y, x))
op_gt = lambda x, y, min, max, inh: max(inh(y, x), x)
op_ge = lambda x, y, min, max, inh: inh(x, inh(x, y))


@torch.jit.script
def t_delta(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    计算TNN中两个脉冲所代表的值（时间）的差值。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 抑制信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = torch.where(torch.isinf(x), torch.full_like(x, torch.inf), torch.where(torch.isinf(y), -torch.full_like(x, torch.inf), x - y))
    return out


def t_to_s(t: torch.Tensor, time_steps: int, t_offset: int = 0) -> torch.Tensor:
    """
    将脉冲时间转换为脉冲序列。
    Args:
        t (torch.Tensor): 时间序列，形状为[...]
        time_steps (int): 最大时间步T
        t_offset (int): 时间步偏移量，从第几个时间步开始，一般为0
    Returns:
        s (torch.Tensor): 脉冲序列，形状为[T, ...]
    """
    return _DF.spike_times_to_spike_train(t, time_steps, t_offset)


@torch.jit.script
def t_rl_forward_add(x: torch.Tensor, t: int) -> torch.Tensor:
    """
    时间序列竞争逻辑的延迟单元前向传播函数。
    Args:
        x (torch.Tensor): 输入信号
        t (int): 延迟时间
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = x + t
    return out


@torch.jit.script
def t_rl_forward_min(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    时间序列竞争逻辑的最小比较单元前向传播函数。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 输入信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = torch.min(x, y)
    return out


@torch.jit.script
def t_rl_forward_max(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    时间序列竞争逻辑的最大比较单元前向传播函数。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 输入信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = torch.max(x, y)
    return out


@torch.jit.script
def t_rl_forward_inh(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    时间序列竞争逻辑的抑制单元前向传播函数。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 抑制信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    delta = t_delta(x, y)
    mask = delta.lt(0.0)
    out = torch.where(mask, x, torch.full_like(x, torch.inf))
    return out


class t_rl_add(torch.autograd.Function):
    @staticmethod
    def forward(ctx: _Any, x: torch.Tensor, t: int) -> torch.Tensor:
        """
        时间序列竞争逻辑延迟单元的前向传播函数。
        Args:
            ctx (Any): 上下文
            x (torch.Tensor): 输入信号
            t (int): 延迟时间
        Returns:
            out (torch.Tensor): 输出信号
        """
        out = t_rl_forward_add(x, t)
        return out
    

    @staticmethod
    def backward(ctx: _Any, grad_output: torch.Tensor) -> torch.Tensor:
        """
        延迟单元的反向传播（伪）函数。
        Args:
            ctx (Any): 上下文
            grad_output (torch.Tensor): 输出信号梯度
        Returns:
            grad_input (torch.Tensor): 输入信号梯度
        """
        return grad_output, None


class t_rl_min(torch.autograd.Function):
    @staticmethod
    def forward(ctx: _Any, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        时间序列竞争逻辑最小比较单元的前向传播函数。
        Args:
            ctx (Any): 上下文
            x (torch.Tensor): 输入信号x
            y (torch.Tensor): 输入信号y
        Returns:
            out (torch.Tensor): 输出信号
        """
        out = t_rl_forward_min(x, y)
        ctx.save_for_backward(t_delta(x, out).eq(0.0), t_delta(y, out).eq(0.0))
        return out
    

    @staticmethod
    def backward(ctx: _Any, grad_output: torch.Tensor) -> torch.Tensor:
        """
        比较单元的反向传播（伪）函数。
        Args:
            ctx (Any): 上下文
            grad_output (torch.Tensor): 输出信号梯度
        Returns:
            grad_x (torch.Tensor): 输入信号x的梯度
            grad_y (torch.Tensor): 输入信号y的梯度
        """
        x_mask, y_mask = ctx.saved_tensors
        x_mask, y_mask = x_mask.to(grad_output), y_mask.to(grad_output)
        grad_x = grad_output * x_mask
        grad_y = grad_output * y_mask
        return grad_x, grad_y


class t_rl_max(t_rl_min):
    @staticmethod
    def forward(ctx: _Any, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        时间序列竞争逻辑最大比较单元的前向传播函数。
        Args:
            ctx (Any): 上下文
            x (torch.Tensor): 输入信号x
            y (torch.Tensor): 输入信号y
        Returns:
            out (torch.Tensor): 输出信号
        """
        out = t_rl_forward_max(x, y)
        ctx.save_for_backward(t_delta(x, out).eq(0.0), t_delta(y, out).eq(0.0))
        return out


class t_rl_inh(torch.autograd.Function):
    @staticmethod
    def forward(ctx: _Any, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        时间序列竞争逻辑抑制单元的前向传播函数。
        Args:
            ctx (Any): 上下文
            x (torch.Tensor): 输入信号x
            y (torch.Tensor): 抑制信号y
        Returns:
            out (torch.Tensor): 输出信号
        """
        out = t_rl_forward_inh(x, y)
        ctx.save_for_backward(t_delta(x, y))
        return out
    

    @staticmethod
    def backward(ctx: _Any, grad_output: torch.Tensor) -> torch.Tensor:
        """
        抑制单元的反向传播（伪）函数。
        Args:
            ctx (Any): 上下文
            grad_output (torch.Tensor): 输出信号梯度
        Returns:
            grad_x (torch.Tensor): 输入信号x的梯度
            grad_y (torch.Tensor): 抑制信号y的梯度
        """
        delta, = ctx.saved_tensors
        mask = ~torch.isinf(delta)
        grad_delta = -_SF.bp_gaussian(delta)
        grad_x = torch.where(mask, grad_delta, torch.zeros_like(grad_output))
        grad_y = -torch.where(mask, grad_delta, torch.zeros_like(grad_output))
        return grad_x, grad_y


def t_add(x: torch.Tensor, t: int) -> torch.Tensor:
    """
    时间序列的延迟运算符。
    Args:
        x (torch.Tensor): 输入信号
        t (int): 延迟时间
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = t_rl_add.apply(x, t)
    return out


def t_min(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    时间序列的最早运算符。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 输入信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = op_min(x, y, t_rl_min.apply, t_rl_max.apply, t_rl_inh.apply)
    return out


def t_xmin(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    时间序列的非同时最早运算符。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 输入信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = op_xmin(x, y, t_rl_min.apply, t_rl_max.apply, t_rl_inh.apply)
    return out


def t_max(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    时间序列的最晚运算符。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 输入信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = op_max(x, y, t_rl_min.apply, t_rl_max.apply, t_rl_inh.apply)
    return out


def t_xmax(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    时间序列的非同时最晚运算符。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 输入信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = op_xmax(x, y, t_rl_min.apply, t_rl_max.apply, t_rl_inh.apply)
    return out


def t_eq(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    时间序列的同时运算符。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 抑制信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = op_eq(x, y, t_rl_min.apply, t_rl_max.apply, t_rl_inh.apply)
    return out


def t_ne(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    时间序列的非同时运算符。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 抑制信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = op_ne(x, y, t_rl_min.apply, t_rl_max.apply, t_rl_inh.apply)
    return out


def t_lt(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    时间序列的早于运算符。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 抑制信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = op_lt(x, y, t_rl_min.apply, t_rl_max.apply, t_rl_inh.apply)
    return out


def t_le(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    时间序列的早于或同时运算符。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 抑制信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = op_le(x, y, t_rl_min.apply, t_rl_max.apply, t_rl_inh.apply)
    return out


def t_gt(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    时间序列的晚于运算符。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 抑制信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = op_gt(x, y, t_rl_min.apply, t_rl_max.apply, t_rl_inh.apply)
    return out


def t_ge(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    时间序列的晚于或同时运算符。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 抑制信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = op_ge(x, y, t_rl_min.apply, t_rl_max.apply, t_rl_inh.apply)
    return out


def s_to_t(s: torch.Tensor) -> torch.Tensor:
    """
    将脉冲序列转换为脉冲时间。若脉冲序列上无脉冲，则将其转为inf。
    Args:
        s (torch.Tensor): 脉冲序列，形状为[T, ...]
    Returns:
        t (torch.Tensor): 时间序列，形状为[...]
    """
    t = _DF.spike_train_to_spike_times(s, torch.inf)
    return t


@torch.jit.script
def s_rl_forward_add(x: torch.Tensor, t: int) -> torch.Tensor:
    """
    脉冲序列竞争逻辑的延迟单元前向传播函数。
    Args:
        x (torch.Tensor): 输入信号
        t (int): 延迟时间
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = torch.zeros_like(x)
    out[t:] = x[:-t]
    return out


@torch.jit.script
def s_rl_forward_min(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    脉冲序列竞争逻辑的最小比较单元前向传播函数。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 输入信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = (x.to(torch.bool) | y.to(torch.bool)).to(x)
    return out


@torch.jit.script
def s_rl_forward_max(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    脉冲序列竞争逻辑的最大比较单元前向传播函数。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 输入信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = (x.to(torch.bool) & y.to(torch.bool)).to(x)
    return out


@torch.jit.script
def s_rl_forward_inh(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    脉冲序列竞争逻辑的抑制单元前向传播函数。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 抑制信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    time_steps = x.shape[0]
    out_seq = []
    s = torch.zeros_like(x[0]).to(torch.bool)
    for t in range(time_steps):
        i = x[t].to(torch.bool)
        j = y[t].to(torch.bool)
        s = (i & ~j) | s
        o = i & s
        out_seq.append(o)
    out = torch.stack(out_seq).to(x)
    return out


class s_rl_add(t_rl_add):
    @staticmethod
    def forward(ctx: _Any, x: torch.Tensor, t: int) -> torch.Tensor:
        """
        脉冲序列竞争逻辑延迟单元的前向传播函数。
        Args:
            ctx (Any): 上下文
            x (torch.Tensor): 输入信号
            t (int): 延迟时间
        Returns:
            out (torch.Tensor): 输出信号
        """
        out = torch.zeros_like(x)
        out[t:] = x[:-t]
        return out


class s_rl_min(t_rl_min):
    @staticmethod
    def forward(ctx: _Any, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        脉冲序列竞争逻辑最小比较单元的前向传播函数。
        Args:
            ctx (Any): 上下文
            x (torch.Tensor): 输入信号x
            y (torch.Tensor): 输入信号y
        Returns:
            out (torch.Tensor): 输出信号
        """
        out = s_rl_forward_min(x, y)
        x_t = s_to_t(x)
        y_t = s_to_t(y)
        out_t = s_to_t(out)
        x_mask = torch.stack([t_delta(x_t, out_t).eq(0.0)] * x.shape[0])
        y_mask = torch.stack([t_delta(y_t, out_t).eq(0.0)] * y.shape[0])
        ctx.save_for_backward(x_mask, y_mask)
        return out


class s_rl_max(t_rl_max):
    @staticmethod
    def forward(ctx: _Any, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        脉冲序列竞争逻辑最大比较单元的前向传播函数。
        Args:
            ctx (Any): 上下文
            x (torch.Tensor): 输入信号x
            y (torch.Tensor): 输入信号y
        Returns:
            out (torch.Tensor): 输出信号
        """
        out = s_rl_forward_max(x, y)
        x_t = s_to_t(x)
        y_t = s_to_t(y)
        out_t = s_to_t(out)
        x_mask = torch.stack([t_delta(x_t, out_t).eq(0.0)] * x.shape[0])
        y_mask = torch.stack([t_delta(y_t, out_t).eq(0.0)] * y.shape[0])
        ctx.save_for_backward(x_mask, y_mask)
        return out


class s_rl_inh(t_rl_inh):
    @staticmethod
    def forward(ctx: _Any, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        脉冲序列竞争逻辑抑制单元的前向传播函数。
        Args:
            ctx (Any): 上下文
            x (torch.Tensor): 输入信号x
            y (torch.Tensor): 抑制信号y
        Returns:
            out (torch.Tensor): 输出信号
        """
        out = s_rl_forward_inh(x, y)
        x_t = s_to_t(x)
        y_t = s_to_t(y)
        delta = torch.stack([t_delta(x_t, y_t)] * x.shape[0])
        ctx.save_for_backward(delta)
        return out


def s_add(x: torch.Tensor, t: int) -> torch.Tensor:
    """
    脉冲序列的延迟运算符。
    Args:
        x (torch.Tensor): 输入信号
        t (int): 延迟时间
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = s_rl_add.apply(x, t)
    return out


def s_min(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    脉冲序列的最早运算符。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 输入信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = op_min(x, y, s_rl_min.apply, s_rl_max.apply, s_rl_inh.apply)
    return out


def s_xmin(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    脉冲序列的非同时最早运算符。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 输入信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = op_xmin(x, y, s_rl_min.apply, s_rl_max.apply, s_rl_inh.apply)
    return out


def s_max(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    脉冲序列的最晚运算符。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 输入信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = op_max(x, y, s_rl_min.apply, s_rl_max.apply, s_rl_inh.apply)
    return out


def s_xmax(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    脉冲序列的非同时最晚运算符。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 输入信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = op_xmax(x, y, s_rl_min.apply, s_rl_max.apply, s_rl_inh.apply)
    return out


def s_eq(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    脉冲序列的同时运算符。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 抑制信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = op_eq(x, y, s_rl_min.apply, s_rl_max.apply, s_rl_inh.apply)
    return out


def s_ne(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    脉冲序列的非同时运算符。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 抑制信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = op_ne(x, y, s_rl_min.apply, s_rl_max.apply, s_rl_inh.apply)
    return out


def s_lt(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    脉冲序列的早于运算符。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 抑制信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = op_lt(x, y, s_rl_min.apply, s_rl_max.apply, s_rl_inh.apply)
    return out


def s_le(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    脉冲序列的早于或同时运算符。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 抑制信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = op_le(x, y, s_rl_min.apply, s_rl_max.apply, s_rl_inh.apply)
    return out


def s_gt(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    脉冲序列的晚于运算符。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 抑制信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = op_gt(x, y, s_rl_min.apply, s_rl_max.apply, s_rl_inh.apply)
    return out


def s_ge(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    脉冲序列的晚于或同时运算符。
    Args:
        x (torch.Tensor): 输入信号x
        y (torch.Tensor): 抑制信号y
    Returns:
        out (torch.Tensor): 输出信号
    """
    out = op_ge(x, y, s_rl_min.apply, s_rl_max.apply, s_rl_inh.apply)
    return out


# -*- coding: UTF-8 -*-
"""
数据变换的操作（类）。
"""


import torch
import torch.nn as nn
from torchvision.transforms import Compose
import matterhorn_pytorch.data.functional as DF
from typing import Iterable


class Clip(nn.Module):
    def __init__(self, clip_range: slice) -> None:
        """
        裁剪事件序列。
        """
        super().__init__()
        self.clip_start = clip_range.start
        self.clip_end = clip_range.stop


    def extra_repr(self) -> str:
        return ", ".join(["start=%d" % self.clip_start, "end=%d" % self.clip_end])


    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x[(x[:, 0] > self.clip_start) & (x[:, 0] < self.clip_end)]


class Sampling(nn.Module):
    def __init__(self, span: slice) -> None:
        """
        对事件序列进行采样。
        """
        super().__init__()
        self.sampling_span = span


    def extra_repr(self) -> str:
        return ", ".join(["span=%d" % self.sampling_span])


    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x[::self.sampling_span]


# -*- coding: UTF-8 -*-
"""
NMNIST数据集。
"""


import numpy as np
import torch
import os
from torchvision.datasets.utils import check_integrity, download_url, extract_archive
from typing import Iterable, Union, Callable, Optional
from urllib.error import URLError
from zipfile import BadZipFile
from rich import print
from rich.progress import track
from matterhorn_pytorch.data.skeleton import EventDataset2d


class NMNIST(EventDataset2d):
    original_data_polarity_exists = True
    original_size = (1, 2, 34, 34)
    mirrors = ["https://data.mendeley.com/public-files/datasets/468j46mzdv/files/"]
    resources = [
        ("39c25547-014b-4137-a934-9d29fa53c7a0/file_downloaded", "Train.zip", "20959b8e626244a1b502305a9e6e2031"),
        ("05a4d654-7e03-4c15-bdfa-9bb2bcbea494/file_downloaded", "Test.zip", "69ca8762b2fe404d9b9bad1103e97832")
    ]
    labels = ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]
    
    
    def __init__(self, root: str, train: bool = True, transform: Optional[Callable] = None, target_transform: Optional[Callable] = None, download: bool = False, cached: bool = True, cache_dtype: torch.dtype = torch.uint8, count: bool = False, time_steps: int = 128, width: int = 34, height: int = 34, polarity: bool = True) -> None:
        """
        NMNIST数据集，将MNIST数据集动态变换后，转为事件的形式。
        Args:
            root (str): 数据集的存储位置
            train (bool): 是否为训练集
            transform (Callable | None): 数据如何变换
            target_transform (Callable | None): 标签如何变换
            download (bool): 如果数据集不存在，是否应该下载
            cached (bool): 是否为数据集作缓存。若为 False，则不作缓存，但是代价是运行速度变慢
            cache_dtype (torch.dtype): 如果为数据集作缓存，缓存的数据类型。默认为8位整型数，若count=True，您可能需要更高的精度储存脉冲张量
            count (bool): 是否采取脉冲计数，若为True则输出张量中各个点脉冲的个数，否则只输出是否有脉冲
            time_steps (int): 最终的数据集总共含有多少个时间步
            width (int): 最终数据集的宽度
            height (int): 最终数据集的高度
            polarity (bool): 最终数据集是否采集极性信息，如果采集，通道数就是2，否则是1
        """
        super().__init__(
            root = root,
            train = train,
            transform = transform,
            target_transform = target_transform,
            download = download,
            cached = cached,
            cache_dtype = cache_dtype,
            count = count,
            t_size = time_steps,
            y_size = height,
            x_size = width,
            polarity = polarity
        )


    @staticmethod
    def filename_to_data(filename: str) -> np.ndarray:
        """
        输入文件名，读取文件内容。
        Args:
            filename (str): 文件名
        Returns:
            data (np.ndarray): 文件内容（数据）
        """
        data_str = ""
        with open(filename, 'rb') as f:
            data_str = f.read()
        data = np.fromstring(data_str, dtype = ">u1")
        return data


    @staticmethod
    def data_to_tpyx(data: np.ndarray) -> np.ndarray:
        """
        将数据分割为t,p,y,x数组。
        Args:
            data (np.ndarray): 数据，形状为[5n]
        Returns:
            data_tpyx (np.ndarray): 分为t,p,y,x的数据，形状为[n, 4]
        """
        res = np.zeros((data.shape[0] // 5, 4), dtype = "uint32")
        res[:, 0] = ((data[2::5] & 0x7f) << 16) + (data[3::5] << 8) + data[4::5]
        res[:, 1] = data[2::5] >> 7
        res[:, 2] = 33 - data[1::5]
        res[:, 3] = data[::5]
        res = res[np.argsort(res[:, 0])]
        return res


    def download(self) -> None:
        """
        下载数据集。
        """
        if self.check_exists():
            return
        os.makedirs(self.raw_folder, exist_ok = True)
        for fileurl, filename, md5 in self.resources:
            if os.path.isfile(os.path.join(self.raw_folder, filename)):
                print("[blue]File %s has already existed.[/blue]" % (os.path.join(self.raw_folder, filename),))
                continue
            is_downloaded = False
            for mirror in self.mirrors:
                url = mirror + fileurl
                try:
                    print("[blue]Downloading %s from %s.[/blue]" % (os.path.join(self.raw_folder, filename), url))
                    download_url(url, root = self.raw_folder, filename = filename, md5 = md5)
                    is_downloaded = True
                    break
                except URLError as error:
                    print("[red]Error in file %s downloaded from %s:\r\n\r\n    %s\r\n\r\nPlease manually download it.[/red]" % (os.path.join(self.raw_folder , filename), url, error))
                    is_downloaded = False
            if is_downloaded:
                print("[green]Successfully downloaded %s.[/green]" % (os.path.join(self.raw_folder, filename),))


    def extract(self) -> None:
        """
        解压下载下来的压缩包。
        """
        zip_file_list = os.listdir(self.raw_folder)
        if not os.path.isdir(self.extracted_folder):
            os.makedirs(self.extracted_folder, exist_ok = True)
        extracted_folder_list = os.listdir(self.extracted_folder)
        if len(zip_file_list) == len(extracted_folder_list):
            print("[blue]Files are already unzipped.[/blue]")
            return
        error_occured = False
        for filename in zip_file_list:
            print("[blue]Trying to unzip file %s ...[/blue]" % (filename,))
            try:
                extract_archive(os.path.join(self.raw_folder, filename), self.extracted_folder)
                print("[green]Sussessfully unzipped file %s.[/green]" % (filename,))
            except BadZipFile as e:
                print("[red]Error in unzipping file %s:\r\n\r\n    %s\r\n\r\nPlease manually fix the problem.[/red]" % (filename, e))
                error_occured = True
        if error_occured:
            raise RuntimeError("There are error(s) in unzipping files.")


    def process(self) -> np.ndarray:
        """
        加载数据集。
        Returns:
            data_label (np.ndarray): 数据信息，包括3列：数据集、标签、其为训练集（1）还是测试集（0）。
        """
        list_filename = os.path.join(self.processed_folder, self.idx_filename)
        if os.path.isfile(list_filename):
            file_list = np.loadtxt(list_filename, dtype = "uint32", delimiter = ",")
            return file_list
        self.extract()
        self.clear_processed()
        os.makedirs(self.processed_folder, exist_ok = True)
        file_list = []
        file_idx = 0
        for is_train in range(2):
            is_train_str = "Train" if is_train else "Test"
            for label in track(range(len(self.labels)), description = "Processing %sing set" % (is_train_str.lower(),)):
                label_str = self.labels[label]
                raw_file_dir = os.path.join(self.extracted_folder, is_train_str, label_str)
                raw_file_list = os.listdir(raw_file_dir)
                for raw_filename in raw_file_list:
                    raw_data = NMNIST.filename_to_data(os.path.join(raw_file_dir, raw_filename))
                    event_data = NMNIST.data_to_tpyx(raw_data)
                    self.create_processed(file_idx, event_data)
                    file_list.append([file_idx, label, is_train])
                    file_idx += 1
        file_list = np.array(file_list, dtype = "uint32")
        np.savetxt(list_filename, file_list, fmt = "%d", delimiter = ",")
        return file_list


# -*- coding: UTF-8 -*-
"""
脉冲神经网络的容器，用来容纳时间和空间维度的脉冲神经网络集合。
建议先在空间维度上构建完整的脉冲神经网络结构，再在多个时间步之内进行模拟。
"""


import torch
import torch.nn as nn
import matterhorn_pytorch.snn.functional as _SF
from matterhorn_pytorch.snn.skeleton import Module as _Module
from typing import Iterable as _Iterable, Mapping as _Mapping, Tuple as _Tuple


def _multi_step_mode_(module: _Module, if_on: bool) -> _Module:
    if if_on and not module.multi_step_mode:
        if module.supports_multi_step_mode:
            module.multi_step_mode_()
        else:
            module = Temporal(module)
    if not if_on and module.multi_step_mode:
        if module.supports_single_step_mode:
            module.single_step_mode_()
        elif isinstance(module, Temporal):
            module = module.module
        else:
            raise ValueError("Unsupported step mode conversion on module %s" % (module.__class__.__name__,))
    return module


class Container(_Module):
    def __init__(self) -> None:
        super().__init__()


class Spatial(Container, nn.Sequential):
    def __init__(self, *args) -> None:
        """
        SNN的空间容器，用法同nn.Sequential，加入一些特殊的作用于SNN的函数。
        Args:
            *args ([nn.Module]): 按空间顺序传入的各个模块
        """
        multi_step_mode = False
        for module in args:
            assert isinstance(module, _Module), "Module %s is not an SNN module." % (module.__class__.__name__)
            multi_step_mode = multi_step_mode or module.multi_step_mode
        Container.__init__(self)
        nn.Sequential.__init__(self, *args)
        self.multi_step_mode_(multi_step_mode)


    def multi_step_mode_(self, if_on: bool = True) -> nn.Module:
        """
        调整模型至多时间步模式。
        Args
            if_on (bool): 当前需要调整为什么模式（True为多时间步模式，False为单时间步模式）
        """
        for idx in range(len(self)):
            module = self[idx]
            self[idx] = _multi_step_mode_(module, if_on)
        return self


    def forward(self, *args, **kwargs) -> torch.Tensor:
        """
        前向传播函数，默认接受的张量形状为[T,B,...]（需要将时间维度通过permute等函数转到最外）
        Args:
            x (torch.Tensor): 输入张量
        Returns:
            y (torch.Tensor): 输出张量
        """
        y = nn.Sequential.forward(self, *args, **kwargs)
        return y


class Temporal(Container):
    def __init__(self, module: _Module) -> None:
        """
        SNN的时间容器，在多个时间步之内执行脉冲神经网络。
        Args:
            module (nn.Module): 所用来执行的单步模型
        """
        assert isinstance(module, _Module) and not module.multi_step_mode, "Temporal container can only accept SNN module with single step mode."
        super().__init__()
        self.multi_step_mode_()
        self.module = module


    @property
    def supports_single_step_mode(self) -> bool:
        """
        是否支持单个时间步。
        Returns:
            if_support (bool): 是否支持单个时间步
        """
        return False


class Sequential(Container, nn.Sequential):
    def __init__(self, *args, multi_step_mode: bool = True) -> None:
        """
        对Sequential进行重写，涵盖ANN与SNN的网络。
        Args:
            args (*nn.Module): 按空间顺序传入的各个模块
        """
        Container.__init__(self)
        nn.Sequential.__init__(self, *args)
        self.multi_step_mode_(multi_step_mode)


    def multi_step_mode_(self, if_on: bool = True) -> nn.Module:
        """
        调整模型至多时间步模式。
        Args
            if_on (bool): 当前需要调整为什么模式（True为多时间步模式，False为单时间步模式）
        """
        for idx in range(len(self)):
            module = self[idx]
            is_snn_module = isinstance(module, _Module)
            if is_snn_module:
                self[idx] = _multi_step_mode_(module, if_on)
        return self


    def forward(self, *args, **kwargs) -> torch.Tensor:
        """
        前向传播函数，默认接受的张量形状为[T,B,...]（需要将时间维度通过permute等函数转到最外）
        Args:
            x (torch.Tensor): 输入张量
        Returns:
            y (torch.Tensor): 输出张量
        """
        y = nn.Sequential.forward(self, *args, **kwargs)
        return y


class ModuleList(Container, nn.ModuleList):
    def __init__(self, modules: _Iterable[nn.Module] = None) -> None:
        Container.__init__(self)
        nn.ModuleList.__init__(
            self,
            modules = modules
        )


class ModuleDict(Container, nn.ModuleDict):
    def __init__(self, modules: _Mapping[str, nn.Module] = None) -> None:
        Container.__init__(self)
        nn.ModuleDict.__init__(
            self,
            modules = modules
        )


class Agent(_Module):
    def __init__(self, nn_module: nn.Module, force_spike_output: bool = False) -> None:
        """
        ANN套壳，用于使ANN模块带有mth.snn.Module的方法
        Args:
            module (nn.Module): torch.nn模型
            force_spike_output (bool): 是否强制脉冲输出
        """
        super().__init__()
        assert not isinstance(nn_module, _Module), "Already an SNN module."
        self.nn_module: nn.Module = nn_module
        self.force_spike_output = force_spike_output


    def multi_step_mode_(self, if_on: bool = True) -> nn.Module:
        """
        调整模型至多时间步模式。
        Args
            if_on (bool): 当前需要调整为什么模式（True为多时间步模式，False为单时间步模式）
        """
        if self.nn_module is not None:
            for name, module in self.nn_module.named_children():
                is_snn_module = isinstance(module, _Module)
                if is_snn_module:
                    setattr(self, name, _multi_step_mode_(module, if_on))
        return self


    def reset(self) -> nn.Module:
        """
        重置模型。
        """
        super().reset()
        if self.nn_module is not None:
            for name, module in self.nn_module.named_children():
                is_snn_module = isinstance(module, _Module)
                if is_snn_module:
                    module.reset()
        return self


    def detach(self) -> nn.Module:
        """
        将模型中的某些变量从其计算图中分离。
        """
        super().detach()
        if self.nn_module is not None:
            for name, module in self.nn_module.named_children():
                is_snn_module = isinstance(module, _Module)
                if is_snn_module:
                    module.detach()
        return self


    def forward_single_time_step(self, *args, **kwargs) -> torch.Tensor:
        """
        单个时间步的前向传播函数。
        Args:
            *args: 输入
            **kwargs: 输入
        Returns:
            res (torch.Tensor): 输出
        """
        if not isinstance(args, _Tuple):
            args = (args,)
        res = self.nn_module(*args, **kwargs)
        return res


    def forward(self, *args, **kwargs) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            *args: 输入
            **kwargs: 输入
        Returns:
            res (torch.Tensor): 输出
        """
        res = super().forward(*args, **kwargs)
        if self.force_spike_output:
            if isinstance(res, _Tuple):
                res = (_SF.val_to_spike(y) if isinstance(y, torch.Tensor) else y for y in res)
            else:
                res = _SF.val_to_spike(res)
        return res


# -*- coding: UTF-8 -*-
"""
TNN的柱，包含兴奋柱、抑制柱与认知柱等。可以看作是组成TNN的层的基本结构。
"""


import torch
import torch.nn as nn
from matterhorn_pytorch.snn import Module as _Module


class Lateral(_Module):
    def __init__(self, tl: int, kl: int) -> None:
        """
        TNN的侧抑制柱
        """
        super().__init__()
        self.tl = tl
        self.kl = kl
    

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        pass


# -*- coding: U_TF-8 -*-
"""
TNN的组件，构建TNN兴奋柱和抑制柱的基本组件。
"""


import torch
import torch.nn as nn
from matterhorn_pytorch.snn import Module as _Module
import matterhorn_pytorch.tnn.functional as _TF


class Bitonic(_Module):
    def __init__(self, level: int, asc: bool = True) -> None:
        """
        双调排序网络。通过时空代数的min和max算子对时空信号进行排序操作。
        Args:
            level (int): 排序的级别l，输入为形状为[T, B, C, 2^l]的脉冲序列
            asc (bool): 是否为升序排序，为True则为升序，否则为降序
        """
        super().__init__()
        self.level = level
        self.asc = asc
        self.length = 2 ** self.level
        self.half_length = 2 ** (self.level - 1)
        if self.level > 1:
            self.asc_unit = Bitonic(self.level - 1, asc = True)
            self.desc_unit = Bitonic(self.level - 1, asc = False)
        else:
            self.asc_unit = None
            self.desc_unit = None
    

    def forward(self, x: torch.Tensor, post: bool = False) -> torch.Tensor:
        """
        双调网络的前向传播函数。
        Args:
            x (torch.Tensor): 输入信号x，形状为[T, B, C, 2^l]
            post (bool): 是否在大排序之后调用，设为False即可
        Returns:
            y (torch.Tensor): 输出信号y，形状为[T, B, C, 2^l]
        """
        if self.level <= 0 or x.shape[-1] <= 1:
            y = x
        elif self.level == 1:
            y0 = _TF.s_min(x[..., 0], x[..., 1])
            y1 = _TF.s_max(x[..., 0], x[..., 1])
            if self.asc:
                y = torch.stack([y0, y1])
            else:
                y = torch.stack([y1, y0])
            y_dims = list(range(len(y.shape)))
            y_dims = y_dims[1:] + [0]
            y = y.permute(*y_dims)
        else:
            if x.shape[-1] <= self.half_length:
                if self.asc:
                    y = self.asc_unit(x, post = True)
                else:
                    y = self.desc_unit(x, post = True)
            else:
                y0 = x[..., :self.half_length]
                y1 = x[..., self.half_length:]
                if not post:
                    y0 = self.asc_unit(y0)
                    y1 = self.desc_unit(y1)
                if self.asc:
                    y2 = _TF.s_min(y0, y1)
                    y3 = _TF.s_max(y0, y1)
                    y0 = self.asc_unit(y2, post = True)
                    y1 = self.asc_unit(y3, post = True)
                else:
                    y2 = _TF.s_max(y0, y1)
                    y3 = _TF.s_min(y0, y1)
                    y0 = self.desc_unit(y2, post = True)
                    y1 = self.desc_unit(y3, post = True)
                y = torch.cat([y0, y1], dim = len(y0.shape) - 1)
        return y


class Firing(_Module):
    def __init__(self, u_threshold: int) -> None:
        """
        统计上升/下降的时间，并发射脉冲。
        """
        super().__init__()
        self.u_threshold = nn.Parameter(torch.tensor(u_threshold), requires_grad = False)
    

    def forward(self, u: torch.Tensor, d: torch.Tensor) -> torch.Tensor:
        """
        统计单元的前向传播函数。
        Args:
            u (torch.Tensor): 上升脉冲序列（按时间由小到大排序），形状为[T, B, C, Q]
            d (torch.Tensor): 下降脉冲序列（按时间由小到大排序），形状为[T, B, C, Q]
        Returns:
            out (torch.Tensor): 结果（是否达到阈值并发射脉冲），形状为[T, B, C, 1]
        """
        q = u.shape[-1]
        upper_threshold = q - self.u_threshold + 1
        u = u[..., -upper_threshold:]
        d = d[..., :upper_threshold]
        res = _TF.s_lt(u, d)
        out = res[..., 0:1]
        for i in range(1, res.shape[-1]):
            out = _TF.s_min(out, res[..., i:i + 1])
        return out


if __name__ == "__main__":
    up_times = Bitonic(4)
    down_times = Bitonic(4)
    firing = Firing(6)
    x = _TF.t_to_s(torch.cat([torch.randint(1, 10, (1, 16)).float()], dim = 1), 16)
    y = _TF.t_to_s(torch.cat([torch.randint(3, 12, (1, 16)).float()], dim = 1), 16)
    print(_TF.s_to_t(x))
    print(_TF.s_to_t(y))
    x = up_times(x)
    y = down_times(y)
    z = firing(x, y)
    print(_TF.s_to_t(z))
    


    




# -*- coding: UTF-8 -*-
"""
Spiking Element-Wise (SEW) ResNet。
Reference:
[Fang W, Yu Z, Chen Y, et al. Deep residual learning in spiking neural networks\[J\]. Advances in Neural Information Processing Systems, 2021, 34: 21056-21069.]
(https://proceedings.neurips.cc/paper/2021/hash/afe434653a898da20044041262b3ac74-Abstract.html)
"""


import torch
import torch.nn as nn
import matterhorn_pytorch.snn as snn
from typing import Tuple, Callable


from matterhorn_pytorch.snn.soma import LIF
if torch.cuda.is_available():
    try:
        from matterhorn_pytorch.snn.soma_cuda import LIF
    except:
        try:
            from matterhorn_pytorch.snn.soma_cpp import LIF
        except:
            pass
else:
    try:
        from matterhorn_pytorch.snn.soma_cpp import LIF
    except:
        pass


class ResADD(snn.Module):
    def __init__(self) -> None:
        super().__init__()


    def forward(self, a: torch.Tensor, s: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数，ADD连接函数。
        Args:
            a (torch.Tensor): 经过两层卷积处理后的脉冲张量
            s (torch.Tensor): 原始脉冲张量
        Returns:
            o (torch.Tensor): 输出脉冲张量
        """
        return a + s


class ResAND(snn.Module):
    def __init__(self) -> None:
        super().__init__()


    def forward(self, a: torch.Tensor, s: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数，AND连接函数。
        Args:
            a (torch.Tensor): 经过两层卷积处理后的脉冲张量
            s (torch.Tensor): 原始脉冲张量
        Returns:
            o (torch.Tensor): 输出脉冲张量
        """
        return a * s


class ResIAND(snn.Module):
    def __init__(self) -> None:
        super().__init__()


    def forward(self, a: torch.Tensor, s: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数，IAND连接函数。
        Args:
            a (torch.Tensor): 经过两层卷积处理后的脉冲张量
            s (torch.Tensor): 原始脉冲张量
        Returns:
            o (torch.Tensor): 输出脉冲张量
        """
        return (1.0 - a) * s


def ConvLIF(in_channels: int, out_channels: int, kernel_size: int = 3, stride: int = 1, tau_m: float = 2.0, spiking_function: snn.Module = snn.Gaussian(), trainable: bool = False):
    """
    绑定卷积神经元（突触）和LIF神经元（胞体）
    Args:
        in_channels (int): 输入脉冲通道数
        out_channels (int): 输出脉冲通道数
        kernel_size (int): 卷积核的大小
        stride (int): 卷积步长
        tau_m (float): 参数τ_{m}，神经元时间常数
        spiking_function (snn.Module): 脉冲函数
        trainable (bool): 参数是否可以训练
    """
    return snn.Sequential(
        snn.Conv2d(
            in_channels = in_channels,
            out_channels = out_channels,
            kernel_size = kernel_size,
            stride = stride,
            padding = kernel_size // 2
        ),
        snn.BatchNorm2d(
            num_features = out_channels
        ),
        LIF(
            tau_m = tau_m,
            spiking_function = spiking_function,
            trainable = trainable
        )
    )


class SEWBlock(snn.Module):
    def __init__(self, in_channels: int, out_channels: int, tau_m: float = 2.0, spiking_function: snn.Module = snn.Gaussian(), residual_connection: snn.Module = ResADD(), down_sampling: bool = False, trainable: bool = False) -> None:
        """
        Spiking Element-Wise Block， SEW ResNet的单元。
        Args:
            in_channels (int): 输入脉冲通道数
            out_channels (int): 输出脉冲通道数
            tau_m (float): 参数τ_{m}，神经元时间常数
            spiking_function (snn.Module): 脉冲函数
            residual_connection (snn.Module): 脉冲连接方式
            down_sampling (bool): 是否进行下采样（出来的图像大小是原大小的一半）
            trainable (bool): 参数是否可以训练
        """
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.conv1 = ConvLIF(
            in_channels = in_channels,
            out_channels = out_channels,
            kernel_size = 3,
            stride = 2 if down_sampling else 1,
            tau_m = tau_m,
            spiking_function = spiking_function,
            trainable = trainable
        )
        self.conv2 = ConvLIF(
            in_channels = out_channels,
            out_channels = out_channels,
            kernel_size = 3,
            stride = 1,
            tau_m = tau_m,
            spiking_function = spiking_function,
            trainable = trainable
        )
        self.down_sampling = down_sampling
        if self.down_sampling:
            self.down_sampling_block = ConvLIF(
                in_channels = in_channels,
                out_channels = out_channels,
                kernel_size = 1,
                stride = 2,
                tau_m = tau_m,
                spiking_function = spiking_function,
                trainable = trainable
            )
        elif in_channels != out_channels:
            self.down_sampling_block = ConvLIF(
                in_channels = in_channels,
                out_channels = out_channels,
                kernel_size = 1,
                tau_m = tau_m,
                spiking_function = spiking_function,
                trainable = trainable
            )
        else:
            self.down_sampling_block = None
        self.residual_connection = residual_connection
    

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 输入脉冲张量
        Returns:
            o (torch.Tensor): 输出脉冲张量
        """
        a = self.conv2(self.conv1(x))
        if self.down_sampling_block is not None:
            s = self.down_sampling_block(x)
        else:
            s = x
        o = self.residual_connection(a, s)
        return o


class SEWRes18(snn.Module):
    def __init__(self, input_h_w: Tuple[int] = (128, 128), num_classes: int = 10, tau_m: float = 2.0, spiking_function: snn.Module = snn.Gaussian(), residual_connection: snn.Module = ResADD(), trainable: bool = False) -> None:
        """
        Spiking Element-Wise Block， SEW ResNet的单元。
        Args:
            input_h_w (int*): 输出脉冲通道数
            num_classes (int): 输出类别数
            tau_m (float): 参数τ_{m}，神经元时间常数
            spiking_function (snn.Module): 脉冲函数
            residual_connection (snn.Module): 脉冲连接方式
            trainable (bool): 参数是否可以训练
        """
        super().__init__()
        self.model = snn.Sequential(
            snn.DirectEncoder(),
            ConvLIF(
                in_channels = 2,
                out_channels = 64,
                kernel_size = 3,
                stride = 2,
                tau_m = tau_m,
                spiking_function = spiking_function,
                trainable = trainable
            ), # 1, [T, 64, 64, 64]
            snn.MaxPool2d(
                kernel_size = 2
            ), # 2, [T, 64, 32, 32]
            SEWBlock(
                in_channels = 64,
                out_channels = 64,
                tau_m = tau_m,
                spiking_function = spiking_function,
                residual_connection = residual_connection,
                trainable = trainable
            ), # 3, 4, [T, 64, 32, 32]
            SEWBlock(
                in_channels = 64,
                out_channels = 64,
                tau_m = tau_m,
                spiking_function = spiking_function,
                residual_connection = residual_connection,
                trainable = trainable
            ), # 5, 6, [T, 64, 32, 32]
            SEWBlock(
                in_channels = 64,
                out_channels = 128,
                tau_m = tau_m,
                spiking_function = spiking_function,
                residual_connection = residual_connection,
                down_sampling = True,
                trainable = trainable
            ), # 7, 8, [T, 128, 16, 16]
            SEWBlock(
                in_channels = 128,
                out_channels = 128,
                tau_m = tau_m,
                spiking_function = spiking_function,
                residual_connection = residual_connection,
                trainable = trainable
            ), # 9, 10, [T, 128, 16, 16]
            SEWBlock(
                in_channels = 128,
                out_channels = 256,
                tau_m = tau_m,
                spiking_function = spiking_function,
                residual_connection = residual_connection,
                down_sampling = True,
                trainable = trainable
            ), # 11, 12, [T, 256, 8, 8]
            SEWBlock(
                in_channels = 256,
                out_channels = 256,
                tau_m = tau_m,
                spiking_function = spiking_function,
                residual_connection = residual_connection,
                trainable = trainable
            ), # 13, 14, [T, 256, 8, 8]
            SEWBlock(
                in_channels = 256,
                out_channels = 512,
                tau_m = tau_m,
                spiking_function = spiking_function,
                residual_connection = residual_connection,
                down_sampling = True,
                trainable = trainable
            ), # 15, 16, [T, 512, 4, 4]
            SEWBlock(
                in_channels = 512,
                out_channels = 512,
                tau_m = tau_m,
                spiking_function = spiking_function,
                residual_connection = residual_connection,
                trainable = trainable
            ), # 17, 18, [T, 512, 4, 4]
            snn.AvgSpikeDecoder(), # [512, 4, 4]
            nn.AvgPool2d(
                kernel_size = (input_h_w[0] >> 5, input_h_w[1] >> 5)
            ), # [512, 1, 1]
            nn.Flatten(), # [512]
            nn.Linear(
                in_features = 512,
                out_features = num_classes
            ), # [10]
            nn.Softmax()
        )


    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 输入脉冲张量
        Returns:
            x (torch.Tensor): 输出张量
        """
        x = self.model(x)
        return x


# -*- coding: UTF-8 -*-
"""
SNN模块的框架，在torch.nn的基础上，定义了几个SNN的基本函数。
"""


import torch
import torch.nn as nn
from typing import Tuple as _Tuple


class Module(nn.Module):
    def __init__(self) -> None:
        """
        脉冲神经网络模块的骨架。
        """
        nn.Module.__init__(self)
        self._multi_step_mode = False


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return ("step_mode=%s" % ('"m"' if self.multi_step_mode else '"s"')) if self.multi_step_mode else ""


    @property
    def supports_single_step_mode(self) -> bool:
        """
        是否支持单个时间步。
        Returns:
            if_support (bool): 是否支持单个时间步
        """
        return True


    @property
    def supports_multi_step_mode(self) -> bool:
        """
        是否支持多个时间步。
        Returns:
            if_support (bool): 是否支持多个时间步
        """
        return True


    @property
    def multi_step_mode(self) -> bool:
        """
        当前是否为多时间步模式。
        Returns:
            if_on (bool): 当前是否为多个时间步模式
        """
        return self._multi_step_mode
    

    @property
    def single_step_mode(self) -> bool:
        """
        当前是否为单时间步模式。
        Returns:
            if_on (bool): 当前是否为单个时间步模式
        """
        return not self._multi_step_mode


    def multi_step_mode_(self, if_on: bool = True) -> nn.Module:
        """
        调整模型至多时间步模式。
        Args
            if_on (bool): 当前需要调整为什么模式（True为多时间步模式，False为单时间步模式）
        """
        if self.supports_multi_step_mode and if_on:
            self._multi_step_mode = True
        elif self.supports_single_step_mode and not if_on:
            self._multi_step_mode = False
        else:
            raise ValueError("Unsupported time step conversion on module %s" % (self.__class__.__name__,))
        return self
    

    def single_step_mode_(self, if_on: bool = True) -> nn.Module:
        """
        调整模型至单时间步模式。
        Args
            if_on (bool): 当前需要调整为什么模式（True为单时间步模式，False为多时间步模式）
        """
        return self.multi_step_mode_(not if_on)


    def reset(self) -> nn.Module:
        """
        重置模型。
        """
        for name, module in self.named_children():
            if isinstance(module, Module):
                module.reset()
        return self


    def detach(self) -> nn.Module:
        """
        将模型中的某些变量从其计算图中分离。
        """
        for name, module in self.named_children():
            if isinstance(module, Module):
                module.detach()
        return self


    def forward_single_time_step(self, *args, **kwargs) -> torch.Tensor:
        """
        单个时间步的前向传播函数。
        Args:
            *args: 输入
            **kwargs: 输入
        Returns:
            res (torch.Tensor): 输出
        """
        pass


    def forward_multi_time_steps(self, *args, **kwargs) -> torch.Tensor:
        """
        多个时间步的前向传播函数。
        Args:
            *args: 输入
            **kwargs: 输入
        Returns:
            res (torch.Tensor): 输出
        """
        if not isinstance(args, _Tuple):
            args = (args,)
        time_steps = args[0].shape[0]
        result = []
        for t in range(time_steps):
            args_t = (x[t] if isinstance(x, torch.Tensor) else x for x in args)
            kwargs_t = {k: x[t] if isinstance(x, torch.Tensor) else x for k, x in kwargs}
            result.append(self.forward_single_time_step(*args_t, **kwargs_t))
        if isinstance(result[0], _Tuple):
            y = (torch.stack([result[t][col] for t in range(len(result))]) if isinstance(result[0][col], torch.Tensor) else result[0][col] for col in range(len(result[0])))
        else:
            y = torch.stack(result)
        return y


    def forward(self, *args, **kwargs) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            *args: 输入
            **kwargs: 输入
        Returns:
            res (torch.Tensor): 输出
        """
        if self.multi_step_mode:
            res = self.forward_multi_time_steps(*args, **kwargs)
        else:
            res = self.forward_single_time_step(*args, **kwargs)
        return res


# -*- coding: UTF-8 -*-
"""
液体状态机的相关函数。
包括邻接矩阵的生成，液体状态机的融合与分离等。
"""


import torch
import torch.nn as nn
from matterhorn_pytorch.lsm.layer import LSM as _LSM
from typing import Iterable


def init_adjacent_direct(neuron_num: int, axon_numbers: Iterable[int], dendrite_numbers: Iterable[int]) -> torch.Tensor:
    """
    自定义连接关系生成邻接矩阵。
    Args:
        neuron_num (int): 液体状态机中神经元的个数
        axon_numbers (int*): 各个轴突（输出）的编号
        dendrite_numbers (int*): 各个树突（输入）的编号
    Returns:
        res (torch.Tensor): 邻接矩阵
    """
    res = torch.zeros(neuron_num, neuron_num, dtype = torch.float)
    res[axon_numbers, dendrite_numbers] = 1.0
    return res


def init_adjacent_uniform(neuron_num: int, threshold: float = 0.75) -> torch.Tensor:
    """
    由均匀分布生成随机邻接矩阵。
    Args:
        neuron_num (int): 液体状态机中神经元的个数
        threshold (float): 连接的阈值，阈值越大代表连接越容易
    Returns:
        res (torch.Tensor): 邻接矩阵
    """
    res = torch.rand(neuron_num, neuron_num)
    return (res <= threshold).to(torch.float)


def init_adjacent_norm(neuron_num: int, threshold: float = 0.75) -> torch.Tensor:
    """
    由正态分布生成随机邻接矩阵。
    Args:
        neuron_num (int): 液体状态机中神经元的个数
        threshold (float): 连接的阈值，阈值越大代表连接越容易
    Returns:
        res (torch.Tensor): 邻接矩阵
    """
    res = torch.randn(neuron_num, neuron_num)
    return (res <= threshold).to(torch.float)


def init_adjacent_dist_1d(length: int, threshold: float = 1.0) -> torch.Tensor:
    """
    由神经元之间的相对（一维）距离得到神经元的邻接矩阵。相对距离越近神经元的连接概率越高。
    Args:
        length (int): 液体状态机中神经元的个数
        threshold (float): 连接的阈值，阈值越大代表连接越容易
    Returns:
        res (torch.Tensor): 邻接矩阵
    """
    neuron_num = length
    res = torch.zeros(neuron_num, neuron_num, dtype = torch.float)
    position = torch.arange(neuron_num).to(torch.float)
    res = torch.abs((res.T + position).T - position)
    res = (res + 1) * torch.rand_like(res)
    return (res <= threshold).to(torch.float)


def init_adjacent_dist_2d(height: int, width: int, threshold: float = 1.0) -> torch.Tensor:
    """
    由神经元之间的相对（二维）距离得到神经元的邻接矩阵。相对距离越近神经元的连接概率越高。
    Args:
        height (int): 二维高度，矩阵中的行数
        width (int): 二维宽度，矩阵中的列数
        threshold (float): 连接的阈值，阈值越大代表连接越容易
    Returns:
        res (torch.Tensor): 邻接矩阵
    """
    neuron_num = height * width
    res = torch.zeros(neuron_num, neuron_num, dtype = torch.float)
    position = torch.arange(neuron_num).to(torch.float)
    position_y = position // width
    position_x = position % width
    res_y = torch.abs((res.T + position_y).T - position_y)
    res_x = torch.abs((res.T + position_x).T - position_x)
    res = torch.sqrt(res_y ** 2 + res_x ** 2)
    res = (res + 1) * torch.rand_like(res)
    return (res <= threshold).to(torch.float)


def init_adjacent_dist_3d(channels: int, height: int, width: int, threshold: float = 1.0) -> torch.Tensor:
    """
    由神经元之间的相对（三维）距离得到神经元的邻接矩阵。相对距离越近神经元的连接概率越高。
    Args:
        channels (int): 三维通道数
        height (int): 三维高度，矩阵中的行数
        width (int): 三维宽度，矩阵中的列数
        threshold (float): 连接的阈值，阈值越大代表连接越容易
    Returns:
        res (torch.Tensor): 邻接矩阵
    """
    neuron_num = channels * height * width
    res = torch.zeros(neuron_num, neuron_num, dtype = torch.float)
    position = torch.arange(neuron_num).to(torch.float)
    position_p = position // (width * height)
    position_yx = position - position_p * width * height
    position_y = position_yx // width
    position_x = position_yx % width
    res_p = torch.abs((res.T + position_p).T - position_p)
    res_y = torch.abs((res.T + position_y).T - position_y)
    res_x = torch.abs((res.T + position_x).T - position_x)
    res = torch.sqrt(res_p ** 2 + res_y ** 2 + res_x ** 2)
    res = (res + 1) * torch.rand_like(res)
    return (res <= threshold).to(torch.float)


# -*- coding: UTF-8 -*-
"""
（未完工）SNN一阶反应函数的表达式解析，可以将任意的一阶反应函数转换至可以被C++和CUDA计算的形式。
"""


import numpy as np
import torch
from typing import List, Tuple, Dict, Union, Any


COMMAND = (
    "NEG" , "ADD" , "SUB" , "MUL" , "DIV" , "NOT" , "AND" , "ORR" ,
    "XOR" , "EQ"  , "NE"  , "LT"  , "LE"  , "GT"  , "GE"  , "JMP" ,
    "POW" , "EXP" , "LOG" , "SIN" , "COS" , "TAN" , "COT" , "SEC" ,
    "CSC" , "ASIN", "ACOS", "ATAN", "ABS" , "SGN" , "RE"  , "IM"  ,
)
SYMBOLS = (
    "!"   , "+"   , "-"   , "*"   , "/"   , "~"   , "&"   , "|"   ,
    "@"   , "=="  , "!="  , "<"   , "<="  , ">"   , ">="  , "?"   ,
    "^"   , "exp" , "ln"  , "sin" , "cos" , "tan" , "cot" , "sec" ,
    "csc" , "asin", "acos", "atan", "abs" , "sgn" , "re"  , "im"  ,
)
OPERAND_TYPE = (
    1     , 2     , 2     , 2     , 2     , 1     , 2     , 2     ,
    2     , 2     , 2     , 2     , 2     , 2     , 2     , 2     ,
    2     , 1     , 1     , 1     , 1     , 1     , 1     , 1     ,
    1     , 1     , 1     , 1     , 1     , 1     , 1     , 1     ,
)
PRIORITY = (
    7     , 3     , 3     , 4     , 4     , 2     , 1     , 1     ,
    1     , 0     , 0     , 0     , 0     , 0     , 0     , 0     ,
    5     , 6     , 6     , 6     , 6     , 6     , 6     , 6     ,
    6     , 6     , 6     , 6     , 6     , 6     , 6     , 6     ,
)


class Node:
    def __init__(self, val = 0, left = None, right = None) -> None:
        self.val = val
        self.left = left
        self.right = right
    

    def __repr__(self) -> str:
        if self.left is None and self.right is None:
            if isinstance(self.val, str):
                return "$%s" % (self.val,)
            elif isinstance(self.val, int):
                return "$%d" % (self.val,)
            elif isinstance(self.val, float):
                return "$%.2f" % (self.val,)
        if OPERAND_TYPE[self.val] == 1:
            return "%s(%s)" % (SYMBOLS[self.val], str(self.left))
        if OPERAND_TYPE[self.val] == 2:
            return "(%s %s %s)" % (str(self.left), SYMBOLS[self.val], str(self.right))
        return ""


def translate_stacks(expr: str):
    bracket_level = 0
    expr_stacks = []
    expr_buffer = ""
    is_operator = lambda s: isinstance(s, str) and (s in SYMBOLS)
    def push_back(es: List, eb: str):
        if eb == "-" and (not len(es) or is_operator(es[-1])):
            eb = "!"
        if len(eb):
            es.append(eb)
            eb = ""
        return es, eb
    
    last_is_op = False
    for idx in range(len(expr)):
        s = expr[idx]
        if s == "(":
            if bracket_level == 0:
                expr_stacks, expr_buffer = push_back(expr_stacks, expr_buffer)
            bracket_level += 1
        if bracket_level > 0:
            expr_buffer += s
        else:
            is_op = is_operator(s)
            if (last_is_op != is_op) or is_op:
                expr_stacks, expr_buffer = push_back(expr_stacks, expr_buffer)
                last_is_op = is_op
            expr_buffer += s
        if s == ")":
            bracket_level -= 1
            if bracket_level == 0:
                expr_stacks, expr_buffer = push_back(expr_stacks, translate_stacks(expr_buffer[1:-1]))
    
    expr_stacks, expr_buffer = push_back(expr_stacks, expr_buffer)
    if bracket_level != 0:
        raise ValueError("You have unclosed brackets. Check your expression.")
    return expr_stacks


def translate_tree_node(node: Any):
    if isinstance(node, Node):
        return node
    if isinstance(node, List):
        return translate_tree(node)
    if node in SYMBOLS:
        raise ValueError("Unknown symbol %s." % (node,))
    return Node(node)


def translate_tree(stacks: List):
    priority_max = max(PRIORITY)
    for p in range(priority_max, -1, -1):
        current_priority_symbols = [symbol if PRIORITY[SYMBOLS.index(symbol)] == p else None for symbol in SYMBOLS]
        idx = 0
        while idx < len(stacks):
            item = stacks[idx]
            if item in current_priority_symbols:
                op = SYMBOLS.index(item)
                if OPERAND_TYPE[op] == 1:
                    assert idx < len(stacks) - 1, "Wrong operand for symbol %s." % (item)
                    operand1 = translate_tree_node(stacks[idx + 1])
                    front = stacks[:idx]
                    rear = stacks[idx + 2:] if idx + 2 < len(stacks) else []
                    stacks = front + [Node(op, operand1)] + rear
                    idx -= 0
                elif OPERAND_TYPE[op] == 2:
                    assert idx > 0 and idx < len(stacks) - 1, "Wrong operand for symbol %s." % (item)
                    operand1 = translate_tree_node(stacks[idx - 1])
                    operand2 = translate_tree_node(stacks[idx + 1])
                    front = stacks[:idx - 1] if idx - 2 >= 0 else []
                    rear = stacks[idx + 2:] if idx + 2 < len(stacks) else []
                    stacks = front + [Node(op, operand1, operand2)] + rear
                    idx -= 1
            idx += 1
    return translate_tree_node(stacks[0])


def translate_variable(var: str) -> float:
    try:
        return float(var)
    except:
        return var


def tree_traverse_update_params(expr_tree: Node, param_list: List = []) -> List:
    fixed_params = ("u", "h", "x")
    if expr_tree.left is not None:
        param_list = tree_traverse_update_params(expr_tree.left, param_list)
    if expr_tree.right is not None:
        param_list = tree_traverse_update_params(expr_tree.right, param_list)
    if expr_tree.left is None and expr_tree.right is None:
        val = translate_variable(expr_tree.val)
        if val in fixed_params:
            expr_tree.val = fixed_params.index(val)
        elif val in param_list:
            expr_tree.val = param_list.index(val) + 3
        else:
            param_list.append(val)
            expr_tree.val = len(param_list) + 2
    return param_list


def translate(expr: str):
    expr_stacks = translate_stacks(expr)
    expr_tree = translate_tree(expr_stacks)
    param_list = tree_traverse_update_params(expr_tree)
    return expr_tree, param_list


def param_init(param_list: List, param_dict: Dict = {}):
    for idx in range(len(param_list)):
        p = param_list[idx]
        if not isinstance(p, float):
            assert p in param_dict, "Unknown parameter %s" % (p,)
            param_list[idx] = param_dict[p]
    return torch.tensor(param_list, dtype = torch.float)


if __name__ == "__main__":
    # a = translate("1+5*exp(x)+3")
    # a = translate("(1.0/tau_m)*(-a_0*(h-u_rest)*(h-u_c)+x)*sin(x)")
    expr_tree, param_list = translate("1.0/tau_m*(-(h-u_rest)+x)")
    print(expr_tree)
    print(param_list)


import os
from subprocess import run
import torch


def cpp_available() -> bool:
    """
    判断C++扩展是否可用。
    Returns:
        is_available (bool): C++扩展是否可用
    """
    try:
        import matterhorn_cpp_extensions as __mth_cppext__
        del __mth_cppext__
        return True
    except:
        return False


def cuda_available() -> bool:
    """
    判断CUDA扩展是否可用。
    Returns:
        is_available (bool): CUDA扩展是否可用
    """
    if not torch.cuda.is_available():
        return False
    try:
        import matterhorn_cuda_extensions as __mth_cuext__
        del __mth_cuext__
        return True
    except:
        return False


def transpose(x: torch.Tensor) -> torch.Tensor:
    """
    转置一个张量。
    Args:
        x (torch.Tensor): 转置前的张量
    Returns:
        x (torch.Tensor): 转置后的张量
    """
    return x.permute(*torch.arange(x.ndim - 1, -1, -1))


# -*- coding: UTF-8 -*-
"""
画图小工具，利用matplotlib库，将数据（事件、电位等）以图片的形式打印下来。
"""


import numpy as np
import torch
import os
import re
import matplotlib
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.font_manager import FontProperties
from typing import Tuple, Iterable, Union, Optional


plt.rcParams['font.family'] = ['SimSun', 'Times New Roman']
plt.rcParams['mathtext.fontset'] = 'stix'


def transition_color(val: float, min_val: float = 0., max_val: float = 1., min_color: str = "#000000", max_color: str = "#ffffff"):
    """
    获取过渡色。
    Args:
        val (float): 当前值
        min_val (float): 最小值
        max_val (float): 最大值
        min_color (str): 当数据处于最小值时，点应当有的颜色
        max_color (str): 当数据处于最大值时，点应当有的颜色
    Returns:
        hex (str): 该值所对应颜色的hex编码
    """
    min_color = "0x" + re.sub(r"[^0-9abcdefABCDEF]", "", min_color)
    max_color = "0x" + re.sub(r"[^0-9abcdefABCDEF]", "", max_color)
    source = int(eval(min_color))
    dest = int(eval(max_color))
    source_r, source_g, source_b = (source >> 16) & 0xff, (source >> 8) & 0xff, source & 0xff
    dest_r, dest_g, dest_b = (dest >> 16) & 0xff, (dest >> 8) & 0xff, dest & 0xff
    delta_r, delta_g, delta_b = dest_r - source_r, dest_g - source_g, dest_b - source_b
    ratio = (val - min_val) / (max_val - min_val)
    target_r, target_g, target_b = int(max(0, min(255, source_r + delta_r * ratio))), int(max(0, min(255, source_g + delta_g * ratio))), int(max(0, min(255, source_b + delta_b * ratio)))
    res = hex((target_r << 16) + (target_g << 8) + target_b).replace("0x", "#")
    if len(res) < 7:
        res = "#" + ("0" * (7 - len(res))) + res[1:]
    return res


def event_plot_1d(indices_pos: np.ndarray, indices_neg: np.ndarray = None, color_pos: str = "#0000ff", color_neg: str = "#ff0000", title: str = None, index_0: str = "t", prec: int = 8) -> None:
    """
    单次一维事件数据打印。
    Args:
        indices_pos (np.ndarray): （正）事件索引
        indices_neg (np.ndarray): 负事件索引，如果没有极性传入None
        color_pos (str): 正事件颜色（Hex）
        color_neg (str): 负事件颜色（Hex）
        title (str): 图片的标题
        index_0 (str): 维度的索引
        prec (int): 事件的脉冲有多细
    """
    if not indices_pos.shape[0]:
        return
    indices_0 = np.arange(prec * (np.max(indices_pos) + 1)) / prec
    indices_1_pos = np.zeros_like(indices_0)
    indices_1_pos[indices_pos * prec] = 1
    plt.plot(indices_0, indices_1_pos, c = color_pos)
    if indices_neg is not None and indices_neg.shape[0]:
        indices_1_neg = np.zeros_like(indices_0)
        indices_1_neg[indices_neg * prec] = 1
        plt.plot(indices_0, indices_1_neg, c = color_neg)
    plt.xlabel(index_0)
    if title is not None:
        plt.title(title)


def event_plot_2d(indices_pos: np.ndarray, indices_neg: np.ndarray = None, values_pos: np.ndarray = None, values_neg: np.ndarray = None, color_pos: str = "#0000ff", color_neg: str = "#ff0000", title: str = None, index_0: str = "t", index_1: str = "x", shape: Tuple = None, shift: float = 0.1) -> None:
    """
    单次二维事件数据打印。
    Args:
        indices_pos (np.ndarray): （正）事件索引
        indices_neg (np.ndarray): 负事件索引，如果没有极性传入None
        values_pos (np.ndarray): 正事件值，默认全为1
        values_pos (np.ndarray): 负事件值，默认全为1
        color_pos (str): 正事件颜色（Hex）
        color_neg (str): 负事件颜色（Hex）
        title (str): 图片的标题
        index_0 (str): 最外维度的索引
        index_1 (str): 最内维度的索引
        shape (Iterable): 画幅的形状
        shift (float): 给事件加上的偏移量，方便同时打印正负事件
    """
    def draw(indices: np.ndarray, values: np.ndarray, color: str, shift: float) -> None:
        if values is None:
            values = np.ones_like(indices)
        min_val = np.min(values)
        max_val = np.max(values)
        if max_val > min_val:
            color_val = ((values - min_val) / (max_val - min_val))
            colors = [transition_color(0.5 + v * 0.5, min_color = "#7f7f7f", max_color = color) for v in color_val]
        else:
            color_val = 1.0
            colors = color
        plt.scatter(indices[:, 0] + shift, indices[:, 1] + shift, s = 1, c = colors)

    if indices_pos.shape[0]:
        draw(indices_pos, values_pos, color_pos, shift)
    if indices_neg is not None and indices_neg.shape[0]:
        draw(indices_neg, values_neg, color_neg, -shift)
    plt.xlabel(index_0)
    plt.ylabel(index_1)
    if shape is not None and len(shape) == 2:
        plt.xlim(-0.5, shape[0] - 0.5)
        plt.ylim(-0.5, shape[1] - 0.5)
    if title is not None:
        plt.title(title)


def spike_train_plot_yx(data: Union[np.ndarray, torch.Tensor], polarity: bool = True, show: bool = True, save: str = None, titles: Iterable[str] = None, figsize: Tuple = (8, 6)) -> None:
    """
    二维空间的脉冲序列数据打印。
    Args:
        data (np.ndarray | torch.Tensor):需要打印的数据（可以是2、3、4维）
        polarity (bool): 是否将第1个维度作为极性维度，如果为True，则传入的数据为[C(P), H, W]（3维）或[B, C(P), H, W]（4维）；如果为False，则传入的数据为[H, W]（2维）或[B, H, W]（3维）
        show (bool): 是否展示图像，默认展示
        save (str): 是否保存图像，若传入路径，则保存图像；否则不保存
        titles (str*): 每张图都是什么标题
        figSize (Tuple): 图像大小
    """
    get_title = lambda b: titles[b] if titles is not None and b >= 0 and b < len(titles) else ("Event Group %d" % (b,))
    if isinstance(data, torch.Tensor):
        data = data.numpy()
    dim = len(data.shape)
    indices = np.array(data.nonzero())
    fig = plt.figure(figsize = figsize)
    if dim == 2: # [H, W]
        event_plot_2d(
            indices[[1, 0]].T,
            values_pos = data[indices[0], indices[1]],
            color_pos = "#00ff00",
            title = get_title(0),
            index_0 = "x",
            index_1 = "y",
            shape = (data.shape[1], data.shape[0])
        )
    if dim == 3:
        if polarity: # [C(P), H, W]
            indices_pos = indices[:, indices[0] == 1][[1, 2]]
            indices_neg = indices[:, indices[0] == 0][[1, 2]]
            event_plot_2d(
                indices_pos[[1, 0]].T,
                indices_neg[[1, 0]].T,
                values_pos = data[1, indices_pos[0], indices_pos[1]],
                values_neg = data[0, indices_neg[0], indices_neg[1]],
                title = get_title(0),
                index_0 = "x",
                index_1 = "y",
                shape = (data.shape[2], data.shape[1])
            )
        else: # [B, H, W]
            batch_size = data.shape[0]
            rows = batch_size
            cols = 1
            if not rows % 2:
                rows = rows // 2
                cols = cols * 2
            for b in range(batch_size):
                indices_this_batch = indices[:, indices[0] == b][[1, 2]]
                plt.subplot(rows, cols, b + 1)
                event_plot_2d(
                    indices_this_batch[[1, 0]].T,
                    values_pos = data[b, indices_this_batch[0], indices_this_batch[1]],
                    color_pos = "#00ff00",
                    title = get_title(b),
                    index_0 = "x",
                    index_1 = "y",
                    shape = (data.shape[2], data.shape[1])
                )
    if dim == 4: # [B, C(P), H, W]
        batch_size = data.shape[0]
        rows = batch_size
        cols = 1
        if not rows % 2:
            rows = rows // 2
            cols = cols * 2
        for b in range(batch_size):
            indices_this_batch = indices[:, indices[0] == b][[1, 2, 3]]
            indices_pos = indices_this_batch[:, indices_this_batch[0] == 1][[1, 2]]
            indices_neg = indices_this_batch[:, indices_this_batch[0] == 0][[1, 2]]
            plt.subplot(rows, cols, b + 1)
            event_plot_2d(
                indices_pos[[1, 0]].T,
                indices_neg[[1, 0]].T,
                values_pos = data[b, 1, indices_pos[0], indices_pos[1]],
                values_neg = data[b, 0, indices_neg[0], indices_neg[1]],
                title = get_title(b),
                index_0 = "x",
                index_1 = "y",
                shape = (data.shape[3], data.shape[2]),
            )
    if show:
        fig.canvas.set_window_title("Event Plotter Result")
        plt.show()
    if save is not None:
        plt.savefig(save)
        plt.close()


def event_seq_plot_yx(data: Union[np.ndarray, torch.Tensor], shape: Iterable, show: bool = True, save: str = None, titles: Iterable[str] = None, figsize: Tuple = (8, 6)) -> None:
    """
    二维空间的事件序列数据打印。
    Args:
        data (np.ndarray | torch.Tensor): 需要打印的数据（可以是2、3维）
        shape (Iterable): 图的形状（长度为data列数的元组），若不指定（或为0）则为数据的长度
        show (bool): 是否展示图像，默认展示
        save (str): 是否保存图像，若传入路径，则保存图像；否则不保存
        titles (str*): 每张图都是什么标题
        figSize (Tuple): 图像大小
    """
    get_title = lambda b: titles[b] if titles is not None and b >= 0 and b < len(titles) else ("Event Group %d" % (b,))
    get_shape = lambda i: shape[i] if shape is not None and len(shape) > i and shape[i] > 0 else (np.max(data[:, i]) - np.min(data[:, i]) + 1)
    if isinstance(data, torch.Tensor):
        data = data.numpy()
    data, values = np.unique(data, axis = 0, return_counts = True)
    dim = data.shape[1]
    fig = plt.figure(figsize = figsize)
    if dim == 3: # [P, H, W]
        indices_pos = data[data[:, 0] == 1][:, [1, 2]]
        indices_neg = data[data[:, 0] == 0][:, [1, 2]]
        values_pos = values[data[:, 0] == 1]
        values_neg = values[data[:, 0] == 0]
        event_plot_2d(
            indices_pos[:, [1, 0]],
            indices_neg[:, [1, 0]],
            values_pos = values_pos,
            values_neg = values_neg,
            title = get_title(0),
            index_0 = "x",
            index_1 = "y",
            shape = (get_shape(2), get_shape(1))
        )
    elif dim == 2: # [H, W]
        event_plot_2d(
            data[:, [1, 0]],
            values_pos = values,
            color_pos = "#00ff00",
            title = get_title(0),
            index_0 = "x",
            index_1 = "y",
            shape = (get_shape(1), get_shape(0))
        )
    if show:
        fig.canvas.set_window_title("Event Plotter Result")
        plt.show()
    if save is not None:
        plt.savefig(save)
        plt.close()


def event_plot_yx(data: Union[np.ndarray, torch.Tensor], shape: Iterable = None, polarity: bool = True, show: bool = True, save: str = None, titles: Iterable[str] = None, figsize: Tuple = (8, 6), is_seq: bool = False) -> None:
    """
    二维空间的事件数据打印。
    Args:
        data (np.ndarray | torch.Tensor): 需要打印的数据
        shape (Iterable): 图的形状（打印序列用）
        polarity (bool): 是否将第1个维度作为极性维度，如果为True，则传入的数据为[C(P), H, W]（3维）或[B, C(P), H, W]（4维）；如果为False，则传入的数据为[H, W]（2维）或[B, H, W]（3维）
        show (bool): 是否展示图像，默认展示
        save (str): 是否保存图像，若传入路径，则保存图像；否则不保存
        titles (str*): 每张图都是什么标题
        figSize (Tuple): 图像大小
        is_seq (bool): 是否为序列，True表明事件为形状为[n, 4]的序列，否则为形状为[T, C(P), H, W]的张量
    """
    if is_seq:
        event_seq_plot_yx(
            data = data,
            shape = shape,
            show = show,
            save = save,
            titles = titles,
            figsize = figsize
        )
    else:
        spike_train_plot_yx(
            data = data,
            polarity = polarity,
            show = show,
            save = save,
            titles = titles,
            figsize = figsize
        )


def spike_train_plot_tx(data: Union[np.ndarray, torch.Tensor], polarity: bool = True, show: bool = True, save: str = None, titles: Iterable[str] = None, figsize: Tuple = (8, 6)) -> None:
    """
    一维空间+一维时间的脉冲序列数据打印。
    Args:
        data (np.ndarray | torch.Tensor):需要打印的数据（可以是2、3、4维）
        polarity (bool): 是否将第2个维度作为极性维度，如果为True，则传入的数据为[T, C(P), L]（3维）或[B, T, C(P), L]（4维）；如果为False，则传入的数据为[T, L]（2维）或[B, T, L]（3维）
        show (bool): 是否展示图像，默认展示
        save (str): 是否保存图像，若传入路径，则保存图像；否则不保存
        titles (str*): 每张图都是什么标题
        figSize (Tuple): 图像大小
    """
    get_title = lambda b: titles[b] if titles is not None and b >= 0 and b < len(titles) else ("Event Group %d" % (b,))
    if isinstance(data, torch.Tensor):
        data = data.numpy()
    dim = len(data.shape)
    indices = np.array(data.nonzero())
    fig = plt.figure(figsize = figsize)
    if dim == 2: # [T, L]
        event_plot_2d(
            indices[[0, 1]].T,
            values_pos = data[indices[0], indices[1]],
            color_pos = "#00ff00",
            title = get_title(0),
            shape = (data.shape[0], data.shape[1])
        )
        event_plot_2d(data, indices, color_pos = "#00ff00", title = get_title(0))
    if dim == 3:
        if polarity: # [T, C(P), L]
            indices_pos = indices[:, indices[1] == 1][[0, 2]]
            indices_neg = indices[:, indices[1] == 0][[0, 2]]
            event_plot_2d(
                indices_pos[[0, 1]].T,
                indices_neg[[0, 1]].T,
                values_pos = data[indices_pos[0], 1, indices_pos[1]],
                values_neg = data[indices_neg[0], 0, indices_neg[1]],
                title = get_title(0),
                shape = (data.shape[0], data.shape[2])
            )
        else: # [B, T, L]
            batch_size = data.shape[0]
            rows = batch_size
            cols = 1
            if not rows % 2:
                rows = rows // 2
                cols = cols * 2
            for b in range(batch_size):
                indices_this_batch = indices[:, indices[0] == b][[1, 2]]
                plt.subplot(rows, cols, b + 1)
                event_plot_2d(
                    indices_this_batch[[0, 1]].T,
                    values_pos = data[b, indices_this_batch[0], indices_this_batch[1]],
                    color_pos = "#00ff00",
                    title = get_title(b),
                    shape = (data.shape[1], data.shape[2])
                )
    if dim == 4: # [B, T, C(P), L]
        batch_size = data.shape[0]
        rows = batch_size
        cols = 1
        if not rows % 2:
            rows = rows // 2
            cols = cols * 2
        for b in range(batch_size):
            indices_this_batch = indices[:, indices[0] == b][[1, 2, 3]]
            indices_pos = indices_this_batch[:, indices_this_batch[1] == 1][[0, 2]]
            indices_neg = indices_this_batch[:, indices_this_batch[1] == 0][[0, 2]]
            plt.subplot(rows, cols, b + 1)
            event_plot_2d(
                indices_pos[[0, 1]].T,
                indices_neg[[0, 1]].T,
                values_pos = data[b, indices_pos[0], 1, indices_pos[1]],
                values_neg = data[b, indices_neg[0], 0, indices_neg[1]],
                title = get_title(b),
                shape = (data.shape[1], data.shape[3])
            )
    if show:
        fig.canvas.set_window_title("Event Plotter Result")
        plt.show()
    if save is not None:
        plt.savefig(save)
        plt.close()


def event_seq_plot_tx(data: Union[np.ndarray, torch.Tensor], shape: Iterable, show: bool = True, save: str = None, titles: Iterable[str] = None, figsize: Tuple = (8, 6)) -> None:
    """
    一维空间+一维时间的事件序列数据打印。
    Args:
        data (np.ndarray | torch.Tensor): 需要打印的数据（可以是2、3维）
        shape (Iterable): 图的形状（长度为data列数的元组），若不指定（或为0）则为数据的长度
        show (bool): 是否展示图像，默认展示
        save (str): 是否保存图像，若传入路径，则保存图像；否则不保存
        titles (str*): 每张图都是什么标题
        figSize (Tuple): 图像大小
    """
    get_title = lambda b: titles[b] if titles is not None and b >= 0 and b < len(titles) else ("Event Group %d" % (b,))
    get_shape = lambda i: shape[i] if shape is not None and len(shape) > i and shape[i] > 0 else (np.max(data[:, i]) - np.min(data[:, i]) + 1)
    if isinstance(data, torch.Tensor):
        data = data.numpy()
    data, values = np.unique(data, axis = 0, return_counts = True)
    dim = data.shape[1]
    fig = plt.figure(figsize = figsize)
    if dim == 3: # [T, C(P), L]
        indices_pos = data[data[:, 1] == 1][:, [0, 2]]
        indices_neg = data[data[:, 1] == 0][:, [0, 2]]
        values_pos = values[data[:, 1] == 1]
        values_neg = values[data[:, 1] == 0]
        event_plot_2d(
            indices_pos[:, [0, 1]],
            indices_neg[:, [0, 1]],
            values_pos = values_pos,
            values_neg = values_neg,
            title = get_title(0),
            shape = (get_shape(0), get_shape(2))
        )
    elif dim == 2: # [T, L]
        event_plot_2d(
            data[:, [0, 1]],
            values_pos = values,
            color_pos = "#00ff00",
            title = get_title(0),
            shape = (get_shape(0), get_shape(1))
        )
    if show:
        fig.canvas.set_window_title("Event Plotter Result")
        plt.show()
    if save is not None:
        plt.savefig(save)
        plt.close()


def event_plot_tx(data: Union[np.ndarray, torch.Tensor], shape: Iterable = None, polarity: bool = True, show: bool = True, save: str = None, titles: Iterable[str] = None, figsize: Tuple = (8, 6), is_seq: bool = False) -> None:
    """
    一维空间+一维时间的事件数据打印。
    Args:
        data (np.ndarray | torch.Tensor): 需要打印的数据
        shape (Iterable): 图的形状（打印序列用）
        polarity (bool): 是否将第2个维度作为极性维度，如果为True，则传入的数据为[T, C(P), L]（3维）或[B, T, C(P), L]（4维）；如果为False，则传入的数据为[T, L]（2维）或[B, T, L]（3维）（打印张量用）
        show (bool): 是否展示图像，默认展示
        save (str): 是否保存图像，若传入路径，则保存图像；否则不保存
        titles (str*): 每张图都是什么标题
        figSize (Tuple): 图像大小
        is_seq (bool): 是否为序列，True表明事件为形状为[n, 4]的序列，否则为形状为[T, C(P), H, W]的张量
    """
    if is_seq:
        event_seq_plot_tx(
            data = data,
            shape = shape,
            show = show,
            save = save,
            titles = titles,
            figsize = figsize
        )
    else:
        spike_train_plot_tx(
            data = data,
            polarity = polarity,
            show = show,
            save = save,
            titles = titles,
            figsize = figsize
        )


def event_plot_3d(ax: Axes3D, indices_pos: np.ndarray, indices_neg: np.ndarray = None, values_pos: np.ndarray = None, values_neg: np.ndarray = None, color_pos: str = "#0000ff", color_neg: str = "#ff0000", title: str = None, index_0: str = "t", index_1: str = "y", index_2: str = "x", shape: Tuple = None, shift: float = 0.1) -> None:
    """
    单次三维事件数据打印。
    Args:
        ax (Axes3D): 画布
        indices_pos (np.ndarray): （正）事件索引
        indices_neg (np.ndarray): 负事件索引，如果没有极性传入None
        values_pos (np.ndarray): 正事件值，默认全为1
        values_pos (np.ndarray): 负事件值，默认全为1
        color_pos (str): 正事件颜色（Hex）
        color_neg (str): 负事件颜色（Hex）
        title (str): 图片的标题
        index_0 (str): 最外维度的索引
        index_1 (str): 中间维度的索引
        index_2 (str): 最内维度的索引
        shape (Iterable): 画幅的形状
        shift (float): 给事件加上的偏移量，方便同时打印正负事件
    """
    def draw(ax: Axes3D, indices: np.ndarray, values: np.ndarray, color: str, shift: float) -> None:
        if values is None:
            values = np.ones_like(indices)
        min_val = np.min(values)
        max_val = np.max(values)
        if max_val > min_val:
            color_val = ((values - min_val) / (max_val - min_val))
            colors = [transition_color(0.5 + v * 0.5, min_color = "#7f7f7f", max_color = color) for v in color_val]
        else:
            color_val = 1.0
            colors = color
        ax.scatter(indices[:, 0] + shift, indices[:, 2] + shift, indices[:, 1] + shift, s = 1, c = colors)
    
    if indices_pos.shape[0]:
        draw(ax, indices_pos, values_pos, color_pos, shift)
    if indices_neg is not None and indices_neg.shape[0]:
        draw(ax, indices_neg, values_neg, color_neg, -shift)
    ax.set_xlabel(index_0)
    ax.set_ylabel(index_2)
    ax.set_zlabel(index_1)
    if shape is not None and len(shape) == 3:
        ax.set_xlim3d(-0.5, shape[0] - 0.5)
        ax.set_ylim3d(-0.5, shape[2] - 0.5)
        ax.set_zlim3d(-0.5, shape[1] - 0.5)
    if title is not None:
        ax.set_title(title)


def spike_train_plot_tyx(data: Union[np.ndarray, torch.Tensor], polarity: bool = True, show: bool = True, save: str = None, titles: Iterable[str] = None, figsize: Tuple = (8, 6)) -> None:
    """
    二维空间+一维时间的脉冲序列数据打印。
    Args:
        data (np.ndarray | torch.Tensor):需要打印的数据（可以是3、4、5维）
        polarity (bool): 是否将第2个维度作为极性维度，如果为True，则传入的数据为[T, C(P), H, W]（4维）或[B, T, C(P), H, W]（5维）；如果为False，则传入的数据为[T, H, W]（3维）或[B, T, H, W]（4维）
        show (bool): 是否展示图像，默认展示
        save (str): 是否保存图像，若传入路径，则保存图像；否则不保存
        titles (str*): 每张图都是什么标题
        figSize (Tuple): 图像大小
    """
    get_title = lambda b: titles[b] if titles is not None and b >= 0 and b < len(titles) else ("Event Group %d" % (b,))
    if isinstance(data, torch.Tensor):
        data = data.numpy()
    dim = len(data.shape)
    indices = np.array(data.nonzero())
    fig = plt.figure(figsize = figsize)
    if dim == 3: # [T, H, W]
        ax = fig.add_subplot(projection='3d')
        event_plot_3d(
            ax,
            indices[[0, 1, 2]].T,
            values_pos = data[indices[0], indices[1], indices[2]],
            color_pos = "#00ff00",
            title = get_title(0),
            shape = (data.shape[0], data.shape[1], data.shape[2])
        )
    if dim == 4:
        if polarity: # [T, C(P), H, W]
            ax = fig.add_subplot(projection='3d')
            indices_pos = indices[:, indices[1] == 1][[0, 2, 3]]
            indices_neg = indices[:, indices[1] == 0][[0, 2, 3]]
            event_plot_3d(
                ax,
                indices_pos[[0, 1, 2]].T,
                indices_neg[[0, 1, 2]].T,
                values_pos = data[indices_pos[0], 1, indices_pos[1], indices_pos[2]],
                values_neg = data[indices_neg[0], 0, indices_neg[1], indices_neg[2]],
                title = get_title(0),
                shape = (data.shape[0], data.shape[2], data.shape[3])
            )
        else: # [B, T, H, W]
            batch_size = data.shape[0]
            rows = batch_size
            cols = 1
            if not rows % 3:
                rows = rows // 3
                cols = cols * 3
            elif not rows % 2:
                rows = rows // 2
                cols = cols * 2
            for b in range(batch_size):
                indices_this_batch = indices[:, indices[0] == b][[1, 2, 3]]
                ax = fig.add_subplot(rows, cols, b + 1, projection='3d')
                event_plot_3d(
                    ax,
                    indices_this_batch[[0, 1, 2]].T,
                    values_pos = data[b, indices_this_batch[0], indices_this_batch[1], indices_this_batch[2]],
                    color_pos = "#00ff00",
                    title = get_title(b),
                    shape = (data.shape[1], data.shape[2], data.shape[3])
                )
                event_plot_3d(ax, data[b], indices[:, indices[0] == b][[1, 2, 3]], color_pos = "#00ff00", title = get_title(b))
    if dim == 5: # [B, T, C(P), H, W]
        batch_size = data.shape[0]
        rows = batch_size
        cols = 1
        if not rows % 3:
            rows = rows // 3
            cols = cols * 3
        elif not rows % 2:
            rows = rows // 2
            cols = cols * 2
        for b in range(batch_size):
            indices_this_batch = indices[:, indices[0] == b][[1, 2, 3, 4]]
            indices_pos = indices_this_batch[:, indices_this_batch[1] == 1][[0, 2, 3]]
            indices_neg = indices_this_batch[:, indices_this_batch[1] == 0][[0, 2, 3]]
            ax = fig.add_subplot(rows, cols, b + 1, projection='3d')
            event_plot_3d(
                ax,
                indices_pos[[0, 1, 2]].T,
                indices_neg[[0, 1, 2]].T,
                values_pos = data[b, indices_pos[0], 1, indices_pos[1], indices_pos[2]],
                values_neg = data[b, indices_neg[0], 0, indices_neg[1], indices_neg[2]],
                title = get_title(b),
                shape = (data.shape[1], data.shape[3], data.shape[4])
            )
    if show:
        fig.canvas.set_window_title("Event Plotter Result")
        plt.show()
    if save is not None:
        plt.savefig(save)
        plt.close()


def event_seq_plot_tyx(data: Union[np.ndarray, torch.Tensor], shape: Iterable, show: bool = True, save: str = None, titles: Iterable[str] = None, figsize: Tuple = (8, 6)) -> None:
    """
    二维空间+一维时间的事件序列数据打印。
    Args:
        data (np.ndarray | torch.Tensor): 需要打印的数据（可以是3、4维）
        shape (Iterable): 图的形状（长度为data列数的元组），若不指定（或为0）则为数据的长度
        show (bool): 是否展示图像，默认展示
        save (str): 是否保存图像，若传入路径，则保存图像；否则不保存
        titles (str*): 每张图都是什么标题
        figSize (Tuple): 图像大小
    """
    get_title = lambda b: titles[b] if titles is not None and b >= 0 and b < len(titles) else ("Event Group %d" % (b,))
    get_shape = lambda i: shape[i] if shape is not None and len(shape) > i and shape[i] > 0 else (np.max(data[:, i]) - np.min(data[:, i]) + 1)
    if isinstance(data, torch.Tensor):
        data = data.numpy()
    data, values = np.unique(data, axis = 0, return_counts = True)
    dim = data.shape[1]
    fig = plt.figure(figsize = figsize)
    ax = fig.add_subplot(projection='3d')
    if dim == 4: # [T, C(P), H, W]
        indices_pos = data[data[:, 1] == 1][:, [0, 2, 3]]
        indices_neg = data[data[:, 1] == 0][:, [0, 2, 3]]
        values_pos = values[data[:, 1] == 1]
        values_neg = values[data[:, 1] == 0]
        event_plot_3d(
            ax,
            indices_pos[:, [0, 1, 2]],
            indices_neg[:, [0, 1, 2]],
            values_pos = values_pos,
            values_neg = values_neg,
            title = get_title(0),
            shape = (get_shape(0), get_shape(2), get_shape(3))
        )
    elif dim == 3: # [T, H, W]
        event_plot_3d(
            ax,
            data[:, [0, 1, 2]],
            values_pos = values,
            color_pos = "#00ff00",
            title = get_title(0),
            shape = (get_shape(0), get_shape(1), get_shape(2))
        )
    if show:
        fig.canvas.set_window_title("Event Plotter Result")
        plt.show()
    if save is not None:
        plt.savefig(save)
        plt.close()


def event_plot_tyx(data: Union[np.ndarray, torch.Tensor], shape: Iterable = None, polarity: bool = True, show: bool = True, save: str = None, titles: Iterable[str] = None, figsize: Tuple = (8, 6), is_seq: bool = False) -> None:
    """
    二维空间+一维时间的事件数据打印。
    Args:
        data (np.ndarray | torch.Tensor): 需要打印的数据
        shape (Iterable): 图的形状（打印序列用）
        polarity (bool): 是否将第2个维度作为极性维度，如果为True，则传入的数据为[T, C(P), H, W]（4维）或[B, T, C(P), H, W]（5维）；如果为False，则传入的数据为[T, H, W]（3维）或[B, T, H, W]（4维）（打印张量用）
        show (bool): 是否展示图像，默认展示
        save (str): 是否保存图像，若传入路径，则保存图像；否则不保存
        titles (str*): 每张图都是什么标题
        figSize (Tuple): 图像大小
        is_seq (bool): 是否为序列，True表明事件为形状为[n, 4]的序列，否则为形状为[T, C(P), H, W]的张量
    """
    if is_seq:
        event_seq_plot_tyx(
            data = data,
            shape = shape,
            show = show,
            save = save,
            titles = titles,
            figsize = figsize
        )
    else:
        spike_train_plot_tyx(
            data = data,
            polarity = polarity,
            show = show,
            save = save,
            titles = titles,
            figsize = figsize
        )


def graph_plot_by_adjacent(adjacent: Union[np.ndarray, torch.Tensor], show: bool = True, save: str = None, figsize: Tuple = (6, 6)) -> None:
    from pyecharts.charts import Graph
    from pyecharts.options import LineStyleOpts
    import webbrowser
    get_name = lambda i: "Neuron %d" % (i + 1,)
    fig = plt.figure(figsize = figsize)
    neuron_num = max(adjacent.shape[0], adjacent.shape[1])
    neurons = []
    for i in range(neuron_num):
        neurons.append({"name": get_name(i), "symbolSize": 10})
    synapses = []
    axons, dendrites = torch.nonzero(adjacent, as_tuple = True)
    axons, dendrites = axons.cpu().numpy().tolist(), dendrites.cpu().numpy().tolist()
    for i in range(len(axons)):
        synapses.append({"source": axons[i], "target": dendrites[i]})
    graph = Graph({
        "title": "LSM Graph",
        "width": "1920px",
        "height": "1080px",
        "is_animation": False
    })
    graph.add(
        "LSM Graph",
        neurons,
        synapses,
        layout = "circular",
        edge_length = 50,
        gravity = 0.2,
        repulsion = 50,
        edge_symbol = [None, "arrow"],
        linestyle_opts = LineStyleOpts(
            curve = 1.0
        ),
    )
    if save:
        graph.render(save)
    else:
        graph.render()
        save = os.path.join(os.getcwd(), "render.html")
    if show:
        webbrowser.open("file://" + save.replace("\\", "/"))


# -*- coding: UTF-8 -*-
"""
脉冲神经网络的整个神经元层，输入为脉冲，输出为脉冲。
由突触将来自上一层神经元的脉冲信号$O_{j}^{l-1}(t)$整合成为突触后电位$X_{i}^{l}(t)$后，在胞体中进行突触后电位的累积和发放。
"""


import torch
import torch.nn as nn
import torch.nn.functional as _F
import matterhorn_pytorch.snn.functional as _SF
from matterhorn_pytorch.snn.skeleton import Module as _Module
from matterhorn_pytorch.snn.soma import Soma as _Soma
from matterhorn_pytorch.snn.container import Temporal as _Temporal
from typing import Any as _Any, Tuple as _Tuple, Iterable as _Iterable, Callable as _Callable, Optional as _Optional, Union as _Union
from torch.nn.common_types import _size_1_t, _size_2_t, _size_3_t, _size_any_t
from torch.types import _size
from matterhorn_pytorch.training.functional import stdp_online as _stdp_online


class Layer(_Module):
    def __init__(self) -> None:
        """
        突触函数的骨架，定义突触最基本的函数。
        """
        super().__init__()


    def forward_multi_time_steps(self, *args, **kwargs) -> torch.Tensor:
        """
        多个时间步的前向传播函数。
        Args:
            *args: 输入
            **kwargs: 输入
        Returns:
            res (torch.Tensor): 输出
        """
        args, kwargs, tb = _SF.merge_time_steps_batch_size(args, kwargs)
        res = self.forward_single_time_step(*args, **kwargs)
        res = _SF.split_time_steps_batch_size(res, tb)
        return res


    def forward(self, *args, **kwargs) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            *args: 输入
            **kwargs: 输入
        Returns:
            res (torch.Tensor): 输出
        """
        res = super().forward(*args, **kwargs)
        if isinstance(res, _Tuple):
            res = (_SF.val_to_spike(y) if isinstance(y, torch.Tensor) else y for y in res)
        else:
            res = _SF.val_to_spike(res)
        return res


class f_stdp_linear(torch.autograd.Function):
    @staticmethod
    def forward(ctx: _Any, input: torch.Tensor, weight: torch.Tensor, input_trace: torch.Tensor, output_trace: torch.Tensor, soma: _Callable, a_pos: float = 0.015, tau_pos: float = 2.0, a_neg: float = 0.015, tau_neg: float = 2.0, training: bool = True, multi_step_mode: bool = True) -> _Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        利用STDP进行学习的全连接层的前向传播函数。
        Args:
            ctx (Any): 上下文
            input (torch.Tensor): 输入脉冲序列
            weight (torch.Tensor): 权重矩阵
            input_trace (torch.Tensor): 输入的迹，累积的输入效应
            output_trace (torch.Tensor): 输出的迹，累积的输出效应
            soma (snn.Module): 胞体
            a_pos (float): STDP参数A+
            tau_pos (float): STDP参数tau+
            a_neg (float): STDP参数A-
            tau_neg (float): STDP参数tau-
            training (bool): 是否正在训练
            multi_step_mode (bool): 是否为多时间步模式
        Returns:
            output (torch.Tensor): 输出脉冲序列
            input_trace (torch.Tensor): 输入的迹，累积的输入效应
            output_trace (torch.Tensor): 输出的迹，累积的输出效应
        """
        if multi_step_mode:
            input_spike_train = input.clone()
            flattened_input, _, tb = _SF.merge_time_steps_batch_size(input)
            time_steps = tb[0]
            psp: torch.Tensor = _F.linear(flattened_input, weight, bias = None)
            psp = _SF.split_time_steps_batch_size(psp, tb)
        else:
            input_spike_train = input[None]
            time_steps = 1
            psp: torch.Tensor = _F.linear(input, weight, bias = None)
        output: torch.Tensor = soma(psp)
        if multi_step_mode:
            output_spike_train = output.clone()
        else:
            output_spike_train = output[None]
        delta_weight = torch.zeros_like(weight)
        if training:
            for t in range(time_steps):
                delta_weight, input_trace, output_trace = _stdp_online(
                    delta_weight = delta_weight, # [O, I]
                    input_trace = input_trace, # [B, O, I]
                    output_trace = output_trace, # [B, O, I]
                    input_spike_train = input_spike_train[t], # [B, I]
                    output_spike_train = output_spike_train[t], # [B, O]
                    a_pos = a_pos,
                    tau_pos = tau_pos,
                    a_neg = a_neg,
                    tau_neg = tau_neg
                )
        ctx.save_for_backward(delta_weight, input)
        return output, input_trace, output_trace


    @staticmethod
    def backward(ctx: _Any, grad_output: torch.Tensor, grad_input_trace: torch.Tensor, grad_output_trace: torch.Tensor) -> _Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, None, None, None, None, None, None, None]:
        """
        利用STDP进行学习的全连接层的反向传播函数。
        Args:
            ctx (Any): 上下文
            grad_output (torch.Tensor): 输出脉冲序列梯度
            grad_input_trace (torch.Tensor): 输入的迹梯度
            grad_output_trace (torch.Tensor): 输出的迹梯度
        Returns:
            grad_input (torch.Tensor): 输入脉冲序列梯度
            grad_weight (torch.Tensor): 权重矩阵梯度
            grad_input_trace (torch.Tensor): 输入的迹梯度
            grad_output_trace (torch.Tensor): 输出的迹梯度
            grad_soma (None): 胞体的梯度，为None
            grad_a_pos (None): STDP参数A+的梯度，为None
            grad_tau_pos (None): STDP参数tau+的梯度，为None
            grad_a_neg (None): STDP参数A-的梯度，为None
            grad_tau_neg (None): STDP参数tau-的梯度，为None
            grad_training (None): 是否正在训练的梯度，为None
            grad_multi_step_mode (None): 是否为多时间步模式的梯度，为None
        """
        delta_weight, input = ctx.saved_tensors
        delta_weight = -delta_weight
        return torch.zeros_like(input), delta_weight, torch.zeros_like(grad_input_trace), torch.zeros_like(grad_output_trace), None, None, None, None, None, None, None


class STDPLinear(_Module):
    def __init__(self, soma: _Soma, in_features: int, out_features: int, a_pos: float = 0.015, tau_pos: float = 2.0, a_neg: float = 0.015, tau_neg: float = 2.0, device: torch.device = None, dtype: torch.dtype = None) -> None:
        """
        使用STDP学习机制时的全连接层。
        Args:
            soma (nn.Module): 使用的脉冲神经元胞体，在matterhorn_pytorch.snn.soma中选择
            in_features (int): 输入长度，用法同nn.Linear
            out_features (int): 输出长度，用法同nn.Linear
            a_pos (float): STDP参数A+
            tau_pos (float): STDP参数tau+
            a_neg (float): STDP参数A-
            tau_neg (float): STDP参数tau-
            device (torch.device): 所计算的设备
            dtype (torch.dtype): 所计算的数据类型
        """
        self.input_trace = None
        self.output_trace = None
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = nn.Parameter(torch.empty((out_features, in_features), device = device, dtype = dtype))
        nn.init.kaiming_uniform_(self.weight, a = 5.0 ** 0.5)
        self.soma = soma
        self.a_pos = a_pos
        self.tau_pos = tau_pos
        self.a_neg = a_neg
        self.tau_neg = tau_neg
        self.reset()


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return ", ".join(["in_features=%d" % self.in_features, "out_features=%d" % self.out_features]) + (", " + Layer.extra_repr(self) if len(Layer.extra_repr(self)) else "")


    def reset(self) -> nn.Module:
        """
        重置模型。
        """
        self.input_trace = _SF.reset_tensor(self.input_trace, 0.0)
        self.output_trace = _SF.reset_tensor(self.output_trace, 0.0)
        return super().reset()


    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 上一层脉冲$O_{j}^{l-1}(t)$
        Returns:
            y (torch.Tensor): 当前层脉冲$O_{i}^{l}(t)$
        """
        if self.multi_step_mode:
            time_steps = x.shape[0]
            batch_size = x.shape[1]
        else:
            time_steps = 1
            batch_size = x.shape[0]
        trace_shape = torch.zeros_like(self.weight)[None].repeat_interleave(batch_size, dim = 0)
        self.input_trace = _SF.init_tensor(self.input_trace, trace_shape)
        self.output_trace = _SF.init_tensor(self.output_trace, trace_shape)
        soma = self.soma.forward_multi_time_steps if self.multi_step_mode else self.soma.forward_single_time_step
        y, self.input_trace, self.output_trace = f_stdp_linear.apply(x, self.weight, self.input_trace, self.output_trace, soma, self.a_pos, self.tau_pos, self.a_neg, self.tau_neg, self.training, self.multi_step_mode)
        return y


class f_stdp_conv2d(torch.autograd.Function):
    @staticmethod
    def forward(ctx: _Any, input: torch.Tensor, weight: torch.Tensor, input_trace: torch.Tensor, output_trace: torch.Tensor, soma: _Callable, stride: _size_any_t, padding: _size_any_t, dilation: _size_any_t, a_pos: float = 0.015, tau_pos: float = 2.0, a_neg: float = 0.015, tau_neg: float = 2.0, training: bool = True, multi_step_mode: bool = True) -> _Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        利用STDP进行学习的2维卷积层的前向传播函数。
        Args:
            ctx (Any): 上下文
            input (torch.Tensor): 输入脉冲序列
            weight (torch.Tensor): 权重矩阵
            input_trace (torch.Tensor): 输入的迹，累积的输入效应
            output_trace (torch.Tensor): 输出的迹，累积的输出效应
            soma (snn.Module): 胞体
            stride (size_t): 卷积的输出步长，决定卷积输出的形状
            padding (size_t): 在边缘填充的量（一般为卷积核大小的一半，向下取整）
            dilation (size_t): 卷积的输入步长
            a_pos (float): STDP参数A+
            tau_pos (float): STDP参数tau+
            a_neg (float): STDP参数A-
            tau_neg (float): STDP参数tau-
            training (bool): 是否正在训练
            multi_step_mode (bool): 是否为多时间步模式
        Returns:
            output (torch.Tensor): 输出脉冲序列
            input_trace (torch.Tensor): 输入的迹，累积的输入效应
            output_trace (torch.Tensor): 输出的迹，累积的输出效应
        """
        if multi_step_mode:
            input_spike_train = input.clone()
            flattened_input, _, tb = _SF.merge_time_steps_batch_size(input)
            time_steps = tb[0]
            psp: torch.Tensor = _F.conv2d(flattened_input, weight, bias = None, stride = tuple(stride), padding = tuple(padding), dilation = tuple(dilation))
            psp = _SF.split_time_steps_batch_size(psp, tb)
        else:
            input_spike_train = input[None]
            time_steps = 1
            psp: torch.Tensor = _F.conv2d(input, weight, bias = None, stride = tuple(stride), padding = tuple(padding), dilation = tuple(dilation))
        output: torch.Tensor = soma(psp)
        if multi_step_mode:
            output_spike_train = output.clone()
        else:
            output_spike_train = output[None]
        delta_weight = torch.zeros_like(weight)
        if training:
            h_in = input_spike_train.shape[3]
            w_in = input_spike_train.shape[4]
            h_out = output_spike_train.shape[3]
            w_out = output_spike_train.shape[4]
            h_wt = weight.shape[2]
            w_wt = weight.shape[3]
            h_stride = stride[0]
            w_stride = stride[1]
            h_padding = padding[0]
            w_padding = padding[1]
            h_dilation = dilation[0]
            w_dilation = dilation[1]
            for y in range(h_out):
                for x in range(w_out):
                    for p in range(h_wt):
                        for q in range(w_wt):
                            u = y * h_stride + p * h_dilation - h_padding
                            v = x * w_stride + q * w_dilation - w_padding
                            if u < 0 or u >= h_in or v < 0 or v >= w_in:
                                continue
                            for t in range(time_steps):
                                delta_weight[:, :, p, q], input_trace[:, :, :, u, v], output_trace[:, :, :, y, x] = _stdp_online(
                                    delta_weight = delta_weight[:, :, p, q], # [CO, CI]
                                    input_trace = input_trace[:, :, :, u, v], # [B, CO, CI]
                                    output_trace = output_trace[:, :, :, y, x], # [B, CO, CI]
                                    input_spike_train = input_spike_train[t, :, :, u, v], # [B, CI]
                                    output_spike_train = output_spike_train[t, :, :, y, x], # [B, CO]
                                    a_pos = a_pos,
                                    tau_pos = tau_pos,
                                    a_neg = a_neg,
                                    tau_neg = tau_neg
                                )
        ctx.save_for_backward(delta_weight, input)
        return output, input_trace, output_trace


    @staticmethod
    def backward(ctx: _Any, grad_output: torch.Tensor, grad_input_trace: torch.Tensor, grad_output_trace: torch.Tensor) -> _Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, None, None, None, None, None, None, None, None, None, None]:
        """
        利用STDP进行学习的2维卷积层的反向传播函数。
        Args:
            ctx (Any): 上下文
            grad_output (torch.Tensor): 输出脉冲序列梯度
            grad_input_trace (torch.Tensor): 输入的迹梯度
            grad_output_trace (torch.Tensor): 输出的迹梯度
        Returns:
            grad_input (torch.Tensor): 输入脉冲序列梯度
            grad_weight (torch.Tensor): 权重矩阵梯度
            grad_input_trace (torch.Tensor): 输入的迹梯度
            grad_output_trace (torch.Tensor): 输出的迹梯度
            grad_soma (None): 胞体的梯度，为None
            grad_stride (size_t): 输出步长的梯度，为None
            grad_padding (size_t): 边缘填充的量的梯度，为None
            grad_dilation (size_t): 输入步长的梯度，为None
            grad_a_pos (None): STDP参数A+的梯度，为None
            grad_tau_pos (None): STDP参数tau+的梯度，为None
            grad_a_neg (None): STDP参数A-的梯度，为None
            grad_tau_neg (None): STDP参数tau-的梯度，为None
            grad_training (None): 是否正在训练的梯度，为None
            grad_multi_step_mode (None): 是否为多时间步模式的梯度，为None
        """
        delta_weight, input = ctx.saved_tensors
        delta_weight = -delta_weight
        return torch.zeros_like(input), delta_weight, torch.zeros_like(grad_input_trace), torch.zeros_like(grad_output_trace), None, None, None, None, None, None, None, None, None, None


class STDPConv2d(_Module):
    def __init__(self, soma: _Soma, in_channels: int, out_channels: int, kernel_size: _size_2_t, stride: _size_2_t = 1, padding: _size_2_t = 0, dilation: _size_2_t = 1, a_pos: float = 0.0002, tau_pos: float = 2.0, a_neg: float = 0.0002, tau_neg: float = 2.0, device: torch.device = None, dtype: torch.dtype = None) -> None:
        """
        使用STDP学习机制时的2维卷积层。
        Args:
            soma (nn.Module): 使用的脉冲神经元胞体，在matterhorn_pytorch.snn.soma中选择
            in_channels (int): 输入的频道数C_{in}
            out_channels (int): 输出的频道C_{out}
            kernel_size (size_2_t): 卷积核的形状
            stride (size_2_t): 卷积的输出步长，决定卷积输出的形状
            padding (size_2_t): 在边缘填充的量（一般为卷积核大小的一半，向下取整）
            dilation (size_2_t): 卷积的输入步长
            a_pos (float): STDP参数A+
            tau_pos (float): STDP参数tau+
            a_neg (float): STDP参数A-
            tau_neg (float): STDP参数tau-
            device (torch.device): 所计算的设备
            dtype (torch.dtype): 所计算的数据类型
        """
        self.input_trace = None
        self.output_trace = None
        super().__init__()
        def _fill(data: _size_any_t, l: int) -> torch.Tensor:
            res = torch.tensor(data)
            if res.ndim == 0:
                res = res[None]
            if res.shape[0] < l:
                res = torch.cat([res] * l)
                res = res[:l]
            return res
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = _fill(kernel_size, 2)
        self.weight = nn.Parameter(torch.empty((out_channels, in_channels, self.kernel_size[0], self.kernel_size[1]), device = device, dtype = dtype))
        nn.init.kaiming_uniform_(self.weight, a = 5.0 ** 0.5)
        self.stride = _fill(stride, 2)
        self.padding = _fill(padding, 2)
        self.dilation = _fill(dilation, 2)
        self.soma = soma
        self.a_pos = a_pos
        self.tau_pos = tau_pos
        self.a_neg = a_neg
        self.tau_neg = tau_neg
        self.reset()


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return ", ".join(["in_channels=%d" % self.in_channels, "out_channels=%d" % self.out_channels, "kernel_size=(%d, %d)" % tuple(self.kernel_size), "stride=(%d, %d)" % tuple(self.stride), "padding=(%d, %d)" % tuple(self.padding), "dilation=(%d, %d)" % tuple(self.dilation)]) + (", " + Layer.extra_repr(self) if len(Layer.extra_repr(self)) else "")


    def reset(self) -> nn.Module:
        """
        重置模型。
        """
        self.input_trace = _SF.reset_tensor(self.input_trace, 0.0)
        self.output_trace = _SF.reset_tensor(self.output_trace, 0.0)
        return super().reset()


    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 上一层脉冲$O_{j}^{l-1}(t)$
        Returns:
            y (torch.Tensor): 当前层脉冲$O_{i}^{l}(t)$
        """
        if self.multi_step_mode:
            time_steps = x.shape[0]
            batch_size = x.shape[1]
            h_in = x.shape[3]
            w_in = x.shape[4]
        else:
            time_steps = 1
            batch_size = x.shape[0]
            h_in = x.shape[2]
            w_in = x.shape[3]
        c_out = self.weight.shape[0]
        c_in = self.weight.shape[1]
        h_wt = self.weight.shape[2]
        w_wt = self.weight.shape[3]
        h_out = (h_in + 2 * self.padding[0] - h_wt * self.dilation[0]) // self.stride[0] + 1
        w_out = (w_in + 2 * self.padding[1] - w_wt * self.dilation[1]) // self.stride[1] + 1
        self.input_trace = _SF.init_tensor(self.input_trace, torch.zeros(batch_size, c_out, c_in, h_in, w_in).to(x))
        self.output_trace = _SF.init_tensor(self.output_trace, torch.zeros(batch_size, c_out, c_in, h_out, w_out))
        soma = self.soma.forward_multi_time_steps if self.multi_step_mode else self.soma.forward_single_time_step
        y, self.input_trace, self.output_trace = f_stdp_conv2d.apply(x, self.weight, self.input_trace, self.output_trace, soma, self.stride, self.padding, self.dilation, self.a_pos, self.tau_pos, self.a_neg, self.tau_neg, self.training, self.multi_step_mode)
        return y


class MaxPool1d(Layer, nn.MaxPool1d):
    def __init__(self, kernel_size: _size_any_t, stride: _Optional[_size_any_t] = None, padding: _size_any_t = 0, dilation: _size_any_t = 1, return_indices: bool = False, ceil_mode: bool = False) -> None:
        """
        一维最大池化。
        Args:
            kernel_size (size_any_t): 池化核大小
            stride (size_any_t | None): 池化步长
            padding (size_any_t): 边界填充的长度
            dilation (size_any_t): 输入侧的池化步长
            return_indices (bool): 是否返回带索引的内容
            ceil_mode (bool): 是否向上取整
        """
        Layer.__init__(self)
        nn.MaxPool1d.__init__(
            self,
            kernel_size = kernel_size,
            stride = stride,
            padding = padding,
            dilation = dilation,
            return_indices = return_indices,
            ceil_mode = ceil_mode
        )


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return nn.MaxPool1d.extra_repr(self) + (", " + Layer.extra_repr(self) if len(Layer.extra_repr(self)) else "")


    def forward_single_time_step(self, x: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 上一层脉冲$O_{j}^{l-1}(t)$
        Returns:
            y (torch.Tensor): 当前层脉冲$O_{i}^{l}(t)$
        """
        y = nn.MaxPool1d.forward(self, x)
        return y


class MaxPool2d(Layer, nn.MaxPool2d):
    def __init__(self, kernel_size: _size_any_t, stride: _Optional[_size_any_t] = None, padding: _size_any_t = 0, dilation: _size_any_t = 1, return_indices: bool = False, ceil_mode: bool = False) -> None:
        """
        二维最大池化。
        Args:
            kernel_size (size_any_t): 池化核大小
            stride (size_any_t | None): 池化步长
            padding (size_any_t): 边界填充的长度
            dilation (size_any_t): 输入侧的池化步长
            return_indices (bool): 是否返回带索引的内容
            ceil_mode (bool): 是否向上取整
        """
        Layer.__init__(self)
        nn.MaxPool2d.__init__(
            self,
            kernel_size = kernel_size,
            stride = stride,
            padding = padding,
            dilation = dilation,
            return_indices = return_indices,
            ceil_mode = ceil_mode
        )


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return nn.MaxPool2d.extra_repr(self) + (", " + Layer.extra_repr(self) if len(Layer.extra_repr(self)) else "")


    def forward_single_time_step(self, x: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 上一层脉冲$O_{j}^{l-1}(t)$
        Returns:
            y (torch.Tensor): 当前层脉冲$O_{i}^{l}(t)$
        """
        y = nn.MaxPool2d.forward(self, x)
        return y


class MaxPool3d(Layer, nn.MaxPool3d):
    def __init__(self, kernel_size: _size_any_t, stride: _Optional[_size_any_t] = None, padding: _size_any_t = 0, dilation: _size_any_t = 1, return_indices: bool = False, ceil_mode: bool = False) -> None:
        """
        三维最大池化。
        Args:
            kernel_size (size_any_t): 池化核大小
            stride (size_any_t | None): 池化步长
            padding (size_any_t): 边界填充的长度
            dilation (size_any_t): 输入侧的池化步长
            return_indices (bool): 是否返回带索引的内容
            ceil_mode (bool): 是否向上取整
        """
        Layer.__init__(self)
        nn.MaxPool3d.__init__(
            self,
            kernel_size = kernel_size,
            stride = stride,
            padding = padding,
            dilation = dilation,
            return_indices = return_indices,
            ceil_mode = ceil_mode
        )


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return nn.MaxPool3d.extra_repr(self) + (", " + Layer.extra_repr(self) if len(Layer.extra_repr(self)) else "")


    def forward_single_time_step(self, x: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 上一层脉冲$O_{j}^{l-1}(t)$
        Returns:
            y (torch.Tensor): 当前层脉冲$O_{i}^{l}(t)$
        """
        y = nn.MaxPool3d.forward(self, x)
        return y


class AvgPool1d(Layer, nn.AvgPool1d):
    def __init__(self, kernel_size: _size_1_t, stride: _Optional[_size_1_t] = None, padding: _size_1_t = 0, ceil_mode: bool = False, count_include_pad: bool = True) -> None:
        """
        一维平均池化。
        Args:
            kernel_size (size_1_t): 池化核大小
            stride (size_1_t): 池化核步长
            padding (size_1_t): 边界填充的长度
            ceil_mode (bool): 是否向上取整
            count_include_pad (bool): 是否连带边界一起计算
        """
        Layer.__init__(self)
        nn.AvgPool1d.__init__(
            self,
            kernel_size = kernel_size,
            stride = stride,
            padding = padding,
            ceil_mode = ceil_mode,
            count_include_pad = count_include_pad
        )


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return nn.AvgPool1d.extra_repr(self) + (", " + Layer.extra_repr(self) if len(Layer.extra_repr(self)) else "")


    def forward_single_time_step(self, x: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 上一层脉冲$O_{j}^{l-1}(t)$
        Returns:
            y (torch.Tensor): 当前层脉冲$O_{i}^{l}(t)$
        """
        y = nn.AvgPool1d.forward(self, x)
        return y


class AvgPool2d(Layer, nn.AvgPool2d):
    def __init__(self, kernel_size: _size_2_t, stride: _Optional[_size_2_t] = None, padding: _size_2_t = 0, ceil_mode: bool = False, count_include_pad: bool = True, divisor_override: _Optional[int] = None) -> None:
        """
        二维平均池化。
        Args:
            kernel_size (size_2_t): 池化核大小
            stride (size_2_t | None): 池化核步长
            padding (size_2_t): 边界填充的长度
            ceil_mode (bool): 是否向上取整
            count_include_pad (bool): 是否连带边界一起计算
            divisor_override (int | None): 是否用某个数取代总和作为除数
        """
        Layer.__init__(self)
        nn.AvgPool2d.__init__(
            self,
            kernel_size = kernel_size,
            stride = stride,
            padding = padding,
            ceil_mode = ceil_mode,
            count_include_pad = count_include_pad,
            divisor_override = divisor_override
        )


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return nn.AvgPool2d.extra_repr(self) + (", " + Layer.extra_repr(self) if len(Layer.extra_repr(self)) else "")


    def forward_single_time_step(self, x: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 上一层脉冲$O_{j}^{l-1}(t)$
        Returns:
            y (torch.Tensor): 当前层脉冲$O_{i}^{l}(t)$
        """
        y = nn.AvgPool2d.forward(self, x)
        return y


class AvgPool3d(Layer, nn.AvgPool3d):
    def __init__(self, kernel_size: _size_3_t, stride: _Optional[_size_3_t] = None, padding: _size_3_t = 0, ceil_mode: bool = False, count_include_pad: bool = True, divisor_override: _Optional[int] = None) -> None:
        """
        三维平均池化。
        Args:
            kernel_size (size_3_t): 池化核大小
            stride (size_3_t | None): 池化核步长
            padding (size_3_t): 边界填充的长度
            ceil_mode (bool): 是否向上取整
            count_include_pad (bool): 是否连带边界一起计算
            divisor_override (int | None): 是否用某个数取代总和作为除数
        """
        Layer.__init__(self)
        nn.AvgPool3d.__init__(
            self,
            kernel_size = kernel_size,
            stride = stride,
            padding = padding,
            ceil_mode = ceil_mode,
            count_include_pad = count_include_pad,
            divisor_override = divisor_override
        )


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return nn.AvgPool3d.extra_repr(self) + (", " + Layer.extra_repr(self) if len(Layer.extra_repr(self)) else "")


    def forward_single_time_step(self, x: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 上一层脉冲$O_{j}^{l-1}(t)$
        Returns:
            y (torch.Tensor): 当前层脉冲$O_{i}^{l}(t)$
        """
        y = nn.AvgPool3d.forward(self, x)
        return y


class MaxUnpool1d(Layer, nn.MaxUnpool1d):
    def __init__(self, kernel_size: _Union[int, _Tuple[int]], stride: _Optional[_Union[int, _Tuple[int]]] = None, padding: _Union[int, _Tuple[int]] = 0) -> None:
        """
        一维最大反池化。
        Args:
            kernel_size (size_3_t): 池化核大小
            stride (size_3_t | None): 池化核步长
            padding (size_3_t): 边界填充的长度
        """
        Layer.__init__(self)
        nn.MaxUnpool1d.__init__(
            self,
            kernel_size = kernel_size,
            stride = stride,
            padding = padding
        )


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return nn.MaxUnpool1d.extra_repr(self) + (", " + Layer.extra_repr(self) if len(Layer.extra_repr(self)) else "")
    

    def forward_single_time_step(self, x: torch.Tensor, indices: torch.Tensor, output_size: _Optional[_Iterable[int]] = None) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 上一层脉冲$O_{j}^{l-1}(t)$
            indices (torch.Tensor): 池化前脉冲的索引
            output_size (torch.Tensor): 输出大小
        Returns:
            y (torch.Tensor): 当前层脉冲$O_{i}^{l}(t)$
        """
        y = nn.MaxUnpool1d.forward(self, x, indices, output_size)
        return y


class MaxUnpool2d(Layer, nn.MaxUnpool2d):
    def __init__(self, kernel_size: _Union[int, _Tuple[int]], stride: _Optional[_Union[int, _Tuple[int]]] = None, padding: _Union[int, _Tuple[int]] = 0) -> None:
        """
        二维最大反池化。
        Args:
            kernel_size (size_3_t): 池化核大小
            stride (size_3_t | None): 池化核步长
            padding (size_3_t): 边界填充的长度
        """
        Layer.__init__(self)
        nn.MaxUnpool2d.__init__(
            self,
            kernel_size = kernel_size,
            stride = stride,
            padding = padding
        )


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return nn.MaxUnpool2d.extra_repr(self) + (", " + Layer.extra_repr(self) if len(Layer.extra_repr(self)) else "")
    

    def forward_single_time_step(self, x: torch.Tensor, indices: torch.Tensor, output_size: _Optional[_Iterable[int]] = None) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 上一层脉冲$O_{j}^{l-1}(t)$
            indices (torch.Tensor): 池化前脉冲的索引
            output_size (torch.Tensor): 输出大小
        Returns:
            y (torch.Tensor): 当前层脉冲$O_{i}^{l}(t)$
        """
        y = nn.MaxUnpool2d.forward(self, x, indices, output_size)
        return y


class MaxUnpool3d(Layer, nn.MaxUnpool3d):
    def __init__(self, kernel_size: _Union[int, _Tuple[int]], stride: _Optional[_Union[int, _Tuple[int]]] = None, padding: _Union[int, _Tuple[int]] = 0) -> None:
        """
        一维最大反池化。
        Args:
            kernel_size (size_3_t): 池化核大小
            stride (size_3_t | None): 池化核步长
            padding (size_3_t): 边界填充的长度
        """
        Layer.__init__(self)
        nn.MaxUnpool3d.__init__(
            self,
            kernel_size = kernel_size,
            stride = stride,
            padding = padding
        )


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return nn.MaxUnpool3d.extra_repr(self) + (", " + Layer.extra_repr(self) if len(Layer.extra_repr(self)) else "")


    def forward_single_time_step(self, x: torch.Tensor, indices: torch.Tensor, output_size: _Optional[_Iterable[int]] = None) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 上一层脉冲$O_{j}^{l-1}(t)$
            indices (torch.Tensor): 池化前脉冲的索引
            output_size (torch.Tensor): 输出大小
        Returns:
            y (torch.Tensor): 当前层脉冲$O_{i}^{l}(t)$
        """
        y = nn.MaxUnpool3d.forward(self, x, indices, output_size)
        return y


class Upsample(nn.Upsample):
    def __init__(self, size: _Optional[_Union[int, _Tuple[int]]] = None, scale_factor: _Optional[_Union[float, _Tuple[float]]] = None, mode: str = 'nearest', align_corners: _Optional[bool] = None, recompute_scale_factor: _Optional[bool] = None) -> None:
        """
        上采样（反池化）。
        Args:
            size (int | int*): 输出大小
            scale_factor (float | float*): 比例因子，如2为上采样两倍
            mode (str): 以何种形式上采样
            align_corners (bool): 若为True，使输入和输出张量的角像素对齐，从而保留这些像素的值
            recompute_scale_factor (bool): 若为True，则必须传入scale_factor并且scale_factor用于计算输出大小。计算出的输出大小将用于推断插值的新比例；若为False，那么size或scale_factor将直接用于插值
        """
        nn.Upsample.__init__(
            self,
            size = size,
            scale_factor = scale_factor,
            mode = mode,
            align_corners = align_corners,
            recompute_scale_factor = recompute_scale_factor
        )


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return nn.Upsample.extra_repr(self) + (", " + Layer.extra_repr(self) if len(Layer.extra_repr(self)) else "")


    def forward_single_time_step(self, x: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 上一层脉冲$O_{j}^{l-1}(t)$
        Returns:
            y (torch.Tensor): 当前层脉冲$O_{i}^{l}(t)$
        """
        y = nn.Upsample.forward(self, x)
        return y


class Flatten(Layer, nn.Flatten):
    def __init__(self, start_dim: int = 2, end_dim: int = -1) -> None:
        """
        展平层。
        Args:
            start_dim (int): 起始维度，默认为2（除去[T, B]之后的维度）
            end_dim (int): 终止维度
        """
        Layer.__init__(self)
        nn.Flatten.__init__(
            self,
            start_dim = max(start_dim - 1, 0) if start_dim >= 0 else start_dim,
            end_dim = max(end_dim - 1, 0) if end_dim >= 0 else end_dim
        )


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return nn.Flatten.extra_repr(self) + (", " + Layer.extra_repr(self) if len(Layer.extra_repr(self)) else "")


    def forward_single_time_step(self, x: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 上一层脉冲$O_{j}^{l-1}(t)$
        Returns:
            y (torch.Tensor): 当前层脉冲$O_{i}^{l}(t)$
        """
        y = nn.Flatten.forward(self, x)
        return y


class Unflatten(Layer, nn.Unflatten):
    def __init__(self, dim: _Union[int, str], unflattened_size: _size) -> None:
        """
        反展开层。
        Args:
            dim (int | str): 在哪个维度反展开
            unflattened_size: 这个维度上的张量要反展开成什么形状
        """
        Layer.__init__(self)
        nn.Unflatten.__init__(
            self,
            dim = max(dim - 1, 0) if dim >= 0 else dim,
            unflattened_size = unflattened_size
        )


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return nn.Unflatten.extra_repr(self) + (", " + Layer.extra_repr(self) if len(Layer.extra_repr(self)) else "")


    def forward_single_time_step(self, x: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 上一层脉冲$O_{j}^{l-1}(t)$
        Returns:
            y (torch.Tensor): 当前层脉冲$O_{i}^{l}(t)$
        """
        y = nn.Unflatten.forward(self, x)
        return y


class Dropout(Layer, nn.Dropout):
    def __init__(self, p: float = 0.5, inplace: bool = False) -> None:
        """
        遗忘层。
        Args:
            p (float): 遗忘概率
            inplace (bool): 是否在原有张量上改动，若为True则直接改原张量，否则新建一个张量
        """
        Layer.__init__(self)
        nn.Dropout.__init__(
            self,
            p = p,
            inplace = inplace
        )


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return nn.Dropout.extra_repr(self) + (", " + Layer.extra_repr(self) if len(Layer.extra_repr(self)) else "")


    def forward_single_time_step(self, x: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 上一层脉冲$O_{j}^{l-1}(t)$
        Returns:
            y (torch.Tensor): 当前层脉冲$O_{i}^{l}(t)$
        """
        y = nn.Dropout.forward(self, x)
        return y


class Dropout1d(Layer, nn.Dropout1d):
    def __init__(self, p: float = 0.5, inplace: bool = False) -> None:
        """
        一维遗忘层。
        Args:
            p (float): 遗忘概率
            inplace (bool): 是否在原有张量上改动，若为True则直接改原张量，否则新建一个张量
        """
        Layer.__init__(self)
        nn.Dropout1d.__init__(
            self,
            p = p,
            inplace = inplace
        )


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return nn.Dropout1d.extra_repr(self) + (", " + Layer.extra_repr(self) if len(Layer.extra_repr(self)) else "")


    def forward_single_time_step(self, x: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 上一层脉冲$O_{j}^{l-1}(t)$
        Returns:
            y (torch.Tensor): 当前层脉冲$O_{i}^{l}(t)$
        """
        y = nn.Dropout1d.forward(self, x)
        return y


class Dropout2d(Layer, nn.Dropout2d):
    def __init__(self, p: float = 0.5, inplace: bool = False) -> None:
        """
        二维遗忘层。
        Args:
            p (float): 遗忘概率
            inplace (bool): 是否在原有张量上改动，若为True则直接改原张量，否则新建一个张量
        """
        Layer.__init__(self)
        nn.Dropout2d.__init__(
            self,
            p = p,
            inplace = inplace
        )


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return nn.Dropout2d.extra_repr(self) + (", " + Layer.extra_repr(self) if len(Layer.extra_repr(self)) else "")


    def forward_single_time_step(self, x: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 上一层脉冲$O_{j}^{l-1}(t)$
        Returns:
            y (torch.Tensor): 当前层脉冲$O_{i}^{l}(t)$
        """
        y = nn.Dropout2d.forward(self, x)
        return y


class Dropout3d(Layer, nn.Dropout3d):
    def __init__(self, p: float = 0.5, inplace: bool = False) -> None:
        """
        三维遗忘层。
        Args:
            p (float): 遗忘概率
            inplace (bool): 是否在原有张量上改动，若为True则直接改原张量，否则新建一个张量
        """
        Layer.__init__(self)
        nn.Dropout3d.__init__(
            self,
            p = p,
            inplace = inplace
        )


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return nn.Dropout3d.extra_repr(self) + (", " + Layer.extra_repr(self) if len(Layer.extra_repr(self)) else "")


    def forward_single_time_step(self, x: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 上一层脉冲$O_{j}^{l-1}(t)$
        Returns:
            y (torch.Tensor): 当前层脉冲$O_{i}^{l}(t)$
        """
        y = nn.Dropout3d.forward(self, x)
        return y